try:
    import torch
except ImportError:
    raise ImportError(
        "\n\nPyTorch is not installed, but it is required for RapidParc.\n" \
        "There are two options: \n"
        "(i)\t For GPU support, please install PyTorch according to your hardware (recommended): \n"
        "\t\t https://pytorch.org/get-started/locally \n"
        "(ii)\t To install the CPU version, run:\n"
        "\t\t pip install rapidparc[cpu]\n"
    )


from .run import RapidParc, RapidParcTckEval
from .test import test
from .train import train

__all__ = ["RapidParc", "RapidParcTckEval", "test", "train"]


import os
from datetime import datetime
import numpy as np
# from concurrent.futures import ThreadPoolExecutor
header = [b'mrtrix tracks\n', b'count: 0000000000\n', b'datatype: Float32LE\n', b'file: . offset\n']

   
class Tck:
    def __init__(self, file_path):
        self.file_path = file_path
        self.parameters = {}
        self.header = []
        self.length = 0
        self.data = None
        self.feature = {}
        self.force = False
        self.f = None

    def get_data_handle(self):
        if self.f is None:
            self.f = open(self.file_path, 'ab')
        return self.f
    
    def add_feature_path(self, feature, path):
        if feature not in self.feature.keys():
            self.feature[feature] = dict(path=path, data=None)
        else:
            print("Feature already existing!")

    def write(self, parameters=None):
        if parameters is None:
            parameters = {}
        if not self.force and os.path.isfile(self.file_path):
            raise Exception('%s exists and shouldnt be overwritten' % self.file_path)
        
        self.parameters = parameters
        self.header = header.copy()
        self.header.append(b'start: %s\n' % bytes(str(datetime.now()), encoding='utf-8'))
        self.header.append(b'end: %s\n' % bytes(str(datetime.now()), encoding='utf-8'))
        for p in self.parameters.keys():
            self.header.append(b'%s: %s\n' % (bytes(p, encoding="utf-8"), bytes(self.parameters[p], encoding="utf-8")))
        self.header.append(b'END\n')
        len_header = len(b''.join(self.header))
        self.header[3] = self.header[3].replace(b'offset', bytes(str(len_header - 6 + len(str(len_header))), encoding='utf-8'))
        with open(self.file_path, 'wb') as f:
            for h in self.header:
                f.write(h)

        for feat in self.feature.keys():
            with open(self.feature[feat]['path'], 'wb') as f:
                for h in self.header:
                    f.write(h)
        if self.data is not None:
            for i in range(self.data.shape[0]):
                self.append(self.data[i], {feat: self.feature[feat]['data'][i] for feat in self.feature.keys()})

    def append(self, path, feature=None, replace=None):
        #print(path)
        v = np.linalg.norm(path)
        if (np.isnan(v) or np.isinf(v)) and not replace:
            return
        if replace:
            path = np.ascontiguousarray(path, dtype='<f4')
        else:
            path = np.ascontiguousarray(np.concatenate([path, [[np.nan, np.nan, np.nan]]]), dtype='<f4')

        # write x,y,z
        self.get_data_handle().write(np.ndarray.tobytes(path))
        # write features in separate files
        if feature:
            for feat in feature.keys():
                path = np.ascontiguousarray(np.concatenate([feature[feat], [np.nan]]), dtype='<f4')
                with open(self.feature[feat]['path'], 'ab') as f:
                    f.write(np.ndarray.tobytes(path))
        self.length += (np.isnan(path)).sum()//3


                        
    def close(self):
        if self.get_data_handle() not in [None, False]:
            self.get_data_handle().close()
        with open(self.file_path, "r+b") as f:
            f.seek(21)
            f.write(b'0' * (10 - len(str(self.length))) + bytes(str(self.length), encoding='utf-8'))
        if self.feature:
            for feat in self.feature.keys():
                with open(self.feature[feat]['path'], 'r+b') as f:
                    f.seek(21)
                    f.write(b'0' * (10 - len(str(self.length))) + bytes(str(self.length), encoding='utf-8'))

        with open(self.file_path, 'ab') as f:
            f.write(np.ndarray.tobytes(np.array([np.inf, np.inf, np.inf], dtype='<f4')))
            # print('File with %s streamlines saved in %s' % (self.length, self.file_path))
            print(f"File saved in {self.file_path}")
        if self.feature:
            for feat in self.feature.keys():
                with open(self.feature[feat]['path'], 'ab') as f:
                    f.write(np.ndarray.tobytes(np.array([np.inf], dtype='<f4')))
                print('File with feature %s saved in %s' % (feat, self.feature[feat]['path']))

    def read(self, min_length=0, resample = None):
        with open(self.file_path, 'rb') as f:
            g = f.readlines()
        g = b''.join(g)
        offset = g.find(b'END') + 4
        d = np.frombuffer(g[offset:], '<f4')
        d = d.reshape((d.shape[0] // 3, 3))
        r = np.hstack([[0], np.where(np.isnan(d[:, 0]))[0]])
        tracts = [d[r[i]:r[i + 1]] for i in range(len(r) - 1) if r[i + 1] - r[i] > min_length]
        self.data = np.array(tracts, dtype=object)
        if resample:
            # Parallelize the loop using ThreadPoolExecutor (or ProcessPoolExecutor)
            self.resampled_data = np.zeros((len(self.data), resample, 3), dtype=np.float32)
            for i, streamline in enumerate(self.data):
                self.resampled_data[i]= streamline[np.linspace(start=0, stop=len(streamline) - 1, num=15).astype(int)]

        if self.feature:
            for feat in self.feature:
                with open(self.feature[feat]['path'], 'rb') as f:
                    g = f.readlines()
                    g = b''.join(g)
                    offset = g.find(b'END') + 4
                    d = np.frombuffer(g[offset:], '<f4')
                    tracts = [d[s] for s in np.ma.clump_unmasked(np.ma.masked_invalid(d))]
                    self.feature[feat]['data'] = np.array(tracts, dtype=object)

        print("File %s read. Contains %s streamlines and %s features" % (self.file_path, self.data.shape[0], len(self.feature.keys())))



import argparse
import os
from .train import train
from datetime import datetime


def main():
    parser = argparse.ArgumentParser(prog="rapidparc-train", description="Argument parser for streamline classification")
    job_id = os.environ.get("WANDB_RUN_ID", None)
    if job_id is None:
        job_id = os.environ.get("SLURM_JOB_ID", default=str(datetime.today().strftime('%Y-%m-%d_%H:%M:%S')))
    

    # learning_params
    parser.add_argument("--slurm_job_id", type=str, default=job_id, help="SLURM_JOB_ID used for identification of a run. Important for elastic restarts.")
    parser.add_argument("--use_ddp", type=bool, default=False, help="If the elastic distributed backend should be used")
    parser.add_argument("--num_epochs", type=int, default=1, help="Number of epochs")
    parser.add_argument("--context_size", type=int, default=2000, help="Context size")
    parser.add_argument("--lr", type=float, default=5e-4, help="Learning rate")
    parser.add_argument("--global_batch_size", type=int, default=8, help="Resulting global batch size after DDP")
    parser.add_argument("--snapshot_every", type=int, default=1000, help="Save snapshot every n epochs")
    parser.add_argument("--store_all_snapshots", type=bool, default=True, help="Store all snapshots")

    # model Params
    parser.add_argument("--embedding_type", type=str, default="flip_aug", help="Can be 'flip_inv' or 'flip_aug'")
    parser.add_argument("--num_layers", type=int, default=1, help="Number of layers")
    parser.add_argument("--d_model", type=int, default=45, help="Model dimension, at least 66 for flip-inv or 45 for flip-aug")
    parser.add_argument("--nhead", type=int, default=1, help="Number of attention heads")
    parser.add_argument("--dim_feedforward", type=int, default=32, help="Feedforward dimension of the transformer")
    parser.add_argument("--dropout", type=float, default=0.1, help="Dropout rate")
    parser.add_argument("--dim_class_hidden", type=int, default=256, help="Hidden dimension for classification")

    # Data Augmentation
    parser.add_argument("--aug_rotate_unif_x", type=list, default=[-45, 45], help="Rotation uniform distribution of x-axis in degrees")
    parser.add_argument("--aug_rotate_unif_y", type=list, default=[-10, 10], help="Rotation uniform distribution of y-axis in degrees")
    parser.add_argument("--aug_rotate_unif_z", type=list, default=[-10, 10], help="Rotation uniform distribution of z-axis in degrees")
    parser.add_argument("--aug_norm_shift_std", type=float, default=0.01, help="Standard deviation of the shift parameter of normalizing layer")
    parser.add_argument("--aug_norm_scale_std", type=float, default=0.01, help="Standard deviation of the scale parameter of normalizing layer")
    parser.add_argument("--aug_noise_sigma", type=float, default=0.001, help="Standard deviation of noise")
    parser.add_argument("--aug_p_hemi", type=float, default=0.3, help="Probability for sampling subtractogram from a single hemisphere")
    parser.add_argument("--aug_method_hemi", type=str, default='random', help="Method for augmenting with a subtractogram hemisphere")

    # Downsampling for limits on input data
    parser.add_argument("--n_vertices", type=int, default=15, help="Sample to this number of vertices per streamline.")

    # Pathing
    parser.add_argument("--out_dir", type=str, default="pretrained_models", help="Output directory for the experiment")
    parser.add_argument("--tractcloud_licence_consent_given", action="store_true", help="Wheter the consent to the tract cloud licence is granted")


    args = parser.parse_args()

    assert args.n_vertices >= 2, "n_vertices must be at least 2"
    assert args.n_vertices <= 15, "n_vertices must be at most 15, this is the maximum resolution of the data provided in TractCloud dataset"

    # Constants:
    args.dim_out = 1600 # 43
    args.sched_args = {
        "scheduler": "CosineAnnealingLR",
        "eta_min": 0.0 # args.lr / 5000
    }
    return train(args=args)


import numpy as np
import torch
import torch.nn as nn
import seaborn as sns
import matplotlib.pyplot as plt
from pathlib import Path
from torch.utils.data import DataLoader
from sklearn.metrics import confusion_matrix, f1_score, accuracy_score, classification_report
import time
from typing import Optional, Union
import os

from .dataset import get_TractCloud_eval_returner




def set_seed(seed: int):
    """
    Set the seed for reproducibility.

    Args:
        seed (int): Seed to set
    """
    torch.manual_seed(seed)
    np.random.seed(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False



def annot_format(val):
    """
    Format the annotations of the confusion matrix such that only values greater than 0 are shown.

    Args:
        val (np.ndarray): Confusion matrix

    Returns:
        np.ndarray: Formatted confusion matrix
    """
    if np.issubdtype(val.dtype, np.integer):
        return np.where(val > 0, np.char.mod('%d', val), '')
    elif np.issubdtype(val.dtype, np.floating):
        return np.where(val > 0.01, np.char.mod('%.2f', val), '')
    else:
        return np.full(val.shape, '')



def create_plot(y_true, y_pred, subject_number, duration, classnames, save_path, unnormalized=False):
    """
    Create a confusion matrix plot.

    Args:
        y_true (np.ndarray): True labels
        y_pred (np.ndarray): Predicted labels
        subject_number (int): Number of the subject
        accuracy (float): Accuracy of the model
        macro_f1_score (float): Macro F1 score of the model
        duration (float): Duration of the prediction
        classnames (np.ndarray): Names of the classes
        save_path (str): Path to save the confusion matrix plot
        unnormalized (bool, optional): Whether to create an unnormalized confusion matrix (default: False)
    """

    accuracy = accuracy_score(y_true=y_true, y_pred=y_pred) * 100
    macro_f1_score = f1_score(y_true=y_true, y_pred=y_pred, average='macro') * 100

    plt.figure(figsize=(25, 20))
    plt.xlabel("Predicted Label")
    plt.ylabel("True Label")
    if subject_number:
        if unnormalized:
            cm = confusion_matrix(y_true=y_true, y_pred=y_pred)
            sns.heatmap(cm, annot=annot_format(cm), xticklabels=classnames, yticklabels=classnames, fmt='s', cmap='Blues', linewidth=.5)
            plt.title(f"Unnormalized Confusion Matrix for Subject {subject_number} \n Accuracy: {accuracy:.2f} %, Macro F1: {macro_f1_score:.2f}, Prediction duration: {duration:.2f} s for {len(y_true)} Streamlines")
            plt.savefig(Path(save_path) / f"confusion_matrix_subject_{subject_number}_unnormalized.pdf")
        else:
            cm = confusion_matrix(y_true=y_true, y_pred=y_pred, normalize='true')
            sns.heatmap(cm, annot=annot_format(cm), xticklabels=classnames, yticklabels=classnames, fmt='s', cmap='Blues', linewidth=.5)
            plt.title(f"Confusion Matrix for Subject {subject_number} \n Accuracy: {accuracy:.2f} %, Macro F1: {macro_f1_score:.2f}, Prediction duration: {duration:.2f} s for {len(y_true)} Streamlines")
            plt.savefig(Path(save_path) / f"confusion_matrix_subject_{subject_number}.pdf")
    else:
        if unnormalized:
            cm = confusion_matrix(y_true=y_true, y_pred=y_pred)
            sns.heatmap(cm, annot=annot_format(cm), xticklabels=classnames, yticklabels=classnames, fmt='s', cmap='Blues', linewidth=.5)
            plt.title(f"Unnormalized Confusion Matrix for all Subjects \n Accuracy: {accuracy:.2f} %, Macro F1: {macro_f1_score:.2f}, Prediction duration: {duration:.2f} s for {len(y_true)} Streamlines")
            plt.savefig(Path(save_path) / f"confusion_matrix_all_subjects_unnormalized.pdf")
        else:
            cm = confusion_matrix(y_true=y_true, y_pred=y_pred, normalize='true')
            sns.heatmap(cm, annot=annot_format(cm), xticklabels=classnames, yticklabels=classnames, fmt='s', cmap='Blues', linewidth=.5)
            plt.title(f"Confusion Matrix for all Subjects \n Accuracy: {accuracy:.2f} %, Macro F1: {macro_f1_score:.2f}, Prediction duration: {duration:.2f} s for {len(y_true)} Streamlines")
            plt.savefig(Path(save_path) / f"confusion_matrix_all_subjects.pdf") 
    plt.close()



@torch.inference_mode()
def create_confusion_matrix(model: nn.Module, 
                            split: str, 
                            input_path: Union[str, os.PathLike], 
                            save_path: Union[str, os.PathLike], 
                            map_800_800_to_43: torch.Tensor, 
                            int_to_label: np.ndarray,
                            eval_context_size: int = 2000, 
                            device: Optional[torch.device] = None, 
                            seed: int = 42, 
                            eval_batch_size: int = 64,
                            n_vertices: Optional[int] = None):
    """
    Create a confusion matrix for a given model and test dataloader.

    Parameters:
        model (torch.nn.Module): 
            The trained model

        split (str): 
            What kind of split - in ["train", "val", "test"]

        input_path (str): 
            Path to the input data
        
        save_path (str): 
            Path to save the confusion matrix plot

        map_800_800_to_43 (torch.tensor): 
            Mapping from 800x800 labels to 43 labels

        eval_context_size (int):
            The context size used for evaluation, assumed to be a devider of 10_000 as TractCloud 
            has 10_000 streamlines per subjects
        
        device (torch.device, optional): 
            Device to use (default: None)

        seed (int):
            random seed, cuda backend is set to be deterministic

        n_vertices (int):
            for downsampling
    """

    set_seed(seed)
    if n_vertices:
        print(f'Streamlines will be downsampled to {n_vertices} points for Confusion plots.')

    if Path(save_path).exists() == False:
        Path(save_path).mkdir(parents=True)

    if device is None:
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    model.to(device)
    model.eval()
    map_800_800_to_43 = map_800_800_to_43.to(device)

    duration = time.time()
    streamlines, labels_43, _ = get_TractCloud_eval_returner(inputPath=input_path, 
                                     device=device, 
                                     eval_context_size=eval_context_size, 
                                     seed=seed,  
                                     n_vertices=n_vertices, 
                                     split=split)()
    prediction_43 = []
    
    # Shapes: [bs, context_size, n_verticies or 15, 3], [bs, context_size], [bs, context_size]
    for i in range(0, len(streamlines), eval_batch_size):
        streamlines_batch = streamlines[i : i + eval_batch_size]
        prediction_800_800_batch = model(streamlines_batch).argmax(dim=-1).flatten() # shape [num_subj * num_streamlines]
        prediction_43.append(map_800_800_to_43[prediction_800_800_batch].cpu().numpy())
    prediction_43 = np.concatenate(prediction_43, axis=0)
    labels_43 = labels_43.flatten().cpu().numpy()  # Shape [num_subj * context_size]

    duration = time.time() - duration


    # Create confusion matrix and classification report
    print(classification_report(y_true=labels_43, y_pred=prediction_43, digits=5, target_names=int_to_label, zero_division=0))

    create_plot(y_true=labels_43, 
                y_pred=prediction_43, 
                subject_number=None, 
                duration=duration, 
                save_path=save_path,
                classnames=int_to_label, 
                unnormalized=False)
    
    create_plot(y_true=labels_43, 
                y_pred=prediction_43, 
                subject_number=None, 
                duration=duration, 
                save_path=save_path,
                classnames=int_to_label, 
                unnormalized=True)
    

def plot_loss_curves(train_losses, val_losses, save_path):
    """
    Plot training and validation loss curves.

    Args:
        train_losses (list): List of training losses
        val_losses (list): List of validation losses
        save_path (str): Path to save the loss curves plot
    """ 
    plt.style.use('seaborn-v0_8')
    fig, ax = plt.subplots(1,2)
    fig.set_size_inches(18,5)

    ax[0].plot(train_losses, c="blue", label="Training Loss", linewidth=3, alpha=0.5)
    ax[0].plot(val_losses, c="red", label="Validation Loss", linewidth=3, alpha=0.5)
    ax[0].legend(loc="best")
    ax[0].set_xlabel("Iteration")
    ax[0].set_ylabel("CE Loss")
    ax[0].set_title("Training Progress (linearscale)")

    ax[1].plot(train_losses, c="blue", label="Training Loss", linewidth=3, alpha=0.5)
    ax[1].plot(val_losses, c="red", label="Validation Loss", linewidth=3, alpha=0.5)
    ax[1].legend(loc="best")
    ax[1].set_xlabel("Iteration")
    ax[1].set_ylabel("CE Loss")
    ax[1].set_yscale("log")
    ax[1].set_title("Training Progress (logscale)")
    plt.savefig(Path(save_path) / "loss_curves.pdf")
    plt.close()


import os
from pathlib import Path
import nibabel as nib
from nibabel.streamlines.array_sequence import ArraySequence
from typing import Union, Tuple
import numpy as np
import torch
import torch.nn as nn
import time
from tqdm import tqdm

from .utils.pypi_package_helper import load_model_and_args_local_or_online, get_mapping_800_800_to_43_online, get_int_to_label_online
from .utils.transforms3D import normalize_to_identity_cube
from .utils import tck_io



def RapidParcTckEval(tck_path: Union[str, os.PathLike],
                     out_path: Union[str, os.PathLike],
                     model_name_or_path: Union[str, os.PathLike],
                     eval_batch_size: int = 512,
                     eval_context_size: int = 2000,
                     device: torch.device = torch.device("cuda" if torch.cuda.is_available() else "cpu"),
                     seed: int = 42,
                     print_time: bool = True,
                     print_class_distribution: bool = True):  
    loaded_streamlines = load_tcks(tck_path=tck_path, verbose=print_time)
    if torch.cuda.is_available():
        device = torch.device("cuda")
    elif torch.backends.mps.is_available():
        device = torch.device("mps")
    else:
        device = torch.device("cpu")
    
    y_pred_43 = RapidParc(model_name_or_path = model_name_or_path,
                          inputTractogram = loaded_streamlines,
                          eval_batch_size=eval_batch_size,
                          eval_context_size=eval_context_size,
                          return_anatomical_clusters=True,
                          return_label_names=False,
                          device=device,
                          seed=seed,
                          print_time=print_time,
                          print_class_distribution=print_class_distribution)
    
    int_to_label = get_int_to_label_online()

    results_to_tck(origStreamlines=loaded_streamlines, 
                   y_pred_43=y_pred_43, 
                   int_to_label=int_to_label,
                   output_path=out_path)
    


def RapidParc(model_name_or_path: Union[str, os.PathLike],
              inputTractogram: Union[list, torch.Tensor, np.ndarray, ArraySequence],
              eval_batch_size: int = 512,
              eval_context_size: int = 2000,
              return_anatomical_clusters: bool = True,
              return_label_names: bool = False,
              device: torch.device = torch.device("cuda" if torch.cuda.is_available() else "cpu"),
              seed: int = 42,
              print_time: bool = False,
              print_class_distribution: bool = False) -> Union[torch.Tensor, np.ndarray]:
    """
    Main evaluation function.

    Parameters:
        model_folder_path (PathLike | str):
            One of the following:
            A pretrained model, i.e. one of "rapidparc", "rapidparc_v8" or "hemiaug"
            Or a path of an experiment run (i.e. a folder of the `pretrained_models` folder after training)

        inputTractogram (list | Tensor | ndarray | ArraySequence):
            The streamlines of a single tractogram. Assumed to be one of the following: 
            NumPy Array of length numStreamlines, where each entry has shape [n_i, 3]
            NumPy Array of shape [numStreamlines, lenStreamline, 3] where lenStreamline >= 15
            Torch Tensor of shape [numStreamlines, lenStreamline, 3] where lenStreamline >= 15
            List of length numStreamlines, where each entry has shape [n_i, 3]
        
        eval_batch_size (int):
            The batch size during evaluation. Scales GPU memory and computational complexity linearly.
            The total number of Streamlines inside one batch = eval_batch_size * eval_context_size

        eval_context_size (int):
            The context size during evaluation. Scales GPU memory and computational complexity quadraticly.
            The total number of Streamlines inside one batch = eval_batch_size * eval_context_size

        return_anatomical_clusters (bool):
            Whether the 43 anatomical clusters should be returned (default). If not, the 800 + 800 clusters
            of the ORG atlas are used.

        return_label_names (bool):
            Whether the cluster names (i.e. "AF", ...) should be returned instead of just the cluster indices.
            This option is only possible if return_anatomical_clusters == True

        device (torch.device):
            Torch device used for evaluation

        seed (int):
            Random seed for reproducible results

        print_time (bool):
            Whether the time measurements should be printed.

        print_class_distribution (bool):
            Whether the class distribution should be printed to command line.

    Returns:
        y_pred (Tensor | ndarray):
            Class predictions in the domain of the 42+1 anatomical regions, 800+800 (ORG atlas) or anatomical cluster names.
    """

    torch.set_float32_matmul_precision('high')
    time_start = time.time()
    model, args = load_model_and_args_local_or_online(name_or_path=model_name_or_path, device=device)
    n_vertices = getattr(args, "n_vertices", None)
    if n_vertices is None:
        n_vertices = getattr(args, "num_support_points_per_streamline", None)
    if n_vertices is None:
        raise AttributeError(
            "Configuration is missing 'n_vertices'. Please add it to args.yaml or provide "
            "'num_support_points_per_streamline'."
        )
    if print_time: print(f"Time to load model: {time.time() - time_start:.2f}s")
    
    # Load Mappings
    mapping_800_800_to_43 = get_mapping_800_800_to_43_online().to(device)
    int_to_label = get_int_to_label_online()

    # Resample Streamlines to n_vertices supporting points if necessary and change datatype to torch.Tensor
    if isinstance(inputTractogram, np.ndarray):
        if inputTractogram.dtype == object or inputTractogram.shape[-1] != n_vertices:
            data_tensor, origStreamlines = resample_number_supporting_points(input_streamlines=inputTractogram, n_vertices=n_vertices, verbose=print_time)
        else:
            origStreamlines = inputTractogram
            data_tensor = torch.from_numpy(inputTractogram).pin_memory()
    elif isinstance(inputTractogram, torch.Tensor):
        if inputTractogram.size(-2) == n_vertices:
            origStreamlines = inputTractogram
            data_tensor = inputTractogram
        else:
            data_tensor, origStreamlines = resample_number_supporting_points(input_streamlines=inputTractogram, n_vertices=n_vertices, verbose=print_time)
    elif isinstance(inputTractogram, list) or isinstance(inputTractogram, ArraySequence):
        data_tensor, origStreamlines = resample_number_supporting_points(input_streamlines=inputTractogram, n_vertices=n_vertices, verbose=print_time)
    else:
        raise ValueError(f"inputTractogram has to be a list, numpy.ndarray or torch.Tensor, but got {type(inputTractogram)}.")
   
    assert data_tensor.shape[1] == n_vertices and data_tensor.shape[2] == 3 and len(data_tensor.shape) == 3, "Assumed the input tensor to have a format of [numStreamlines, n_vertices, 3]"

    # Normalize Streamlines to the [-1, 1] cube
    data_tensor = normalize_to_identity_cube(data_tensor.unsqueeze(0)).squeeze(0)

    # Shuffle data + Split in context size
    rng = torch.Generator(device="cpu").manual_seed(seed)
    time_to_shuffle_data = time.time()
    data_tensor, origStreamlines, permutation_indeces = shuffle_and_split_in_context_size(data_tensor=data_tensor, 
                                                                    origStreamlines=origStreamlines, 
                                                                    context_size=eval_context_size,
                                                                    rng=rng)
    if print_time: 
        print(f"Time to shuffle data: {time.time() - time_to_shuffle_data:.2f}s")
        print(f"{len(origStreamlines)} Streamlines are loaded and preprocessed.")

    # Forward Pass
    start_eval_time = time.time()
    y_pred = model_forward_pass(model=model, 
                                   eval_batch_size=eval_batch_size,
                                   data_tensor=data_tensor,
                                   mapping_800_800_to_43=mapping_800_800_to_43,
                                   device=device, 
                                   return_anatomical_clusters=return_anatomical_clusters)
    
    duration_eval_time = time.time() - start_eval_time
    if print_time: print(f"Prediction is done in {duration_eval_time:.2f}s")

    # Shorten Result-tensor to the original length
    y_pred = y_pred[:len(origStreamlines)]

    # Undo the permutation
    y_pred = y_pred[torch.argsort(permutation_indeces)]

    ## Create Histogram
    if print_class_distribution:
        print_command_line_bar_plot(y_pred=y_pred.numpy(), int_to_label=int_to_label)

    if print_time: 
        print(f"Total time: {time.time() - time_start:.2f}s")
    
    if return_anatomical_clusters and return_label_names:
        return int_to_label[y_pred.numpy()]
    return y_pred



@torch.inference_mode()
def resample_number_supporting_points(input_streamlines: Union[list, np.ndarray, torch.Tensor], 
                                      n_vertices: int,
                                      verbose: bool = False) -> Tuple[torch.Tensor, np.ndarray]:
    """
    Function that resamples the number of supporting points (n_vertices).

    Args:
        input_streamlines (list): List of streamlines. Each streamline is a numpy array of shape [N_i, 3].
        verbose (bool): If True, print the time it took to shorten the streamlines.

    Returns:
        resampled_streamlines (torch.Tensor): The resampled streamlines. Shape [N, n_vertices, 3].
        origStreamlines (np.ndarray): The original streamlines as a numpy array object (for slicing operations).
    """
    time_to_shorten_streamlines = time.time()

    origStreamlines = []
    resampled_streamlines = []
    # Select all Streamlines with length ≥ n_vertices
    for i, streamline in enumerate(input_streamlines):
        if len(streamline) < n_vertices:
            print(f"""Streamline at index {i} has a length {len(streamline)}, which is less than {n_vertices}.
                  As a result, the classification result might be faulty. 
                  Make sure to use streamlines with at least {n_vertices} supporting points and that the shape
                  of each streamline is [N_i, 3].""")
        assert len(streamline[0]) == 3, f"Streamline at index {i} has a shape {streamline[0].shape}, but assumed [N, 3]."
        origStreamlines.append(streamline)
        indices = np.round(np.linspace(start=0, stop=len(streamline) - 1, num=n_vertices)).astype(int)
        selected_points = streamline[indices]
        resampled_streamlines.append(selected_points)
    resampled_streamlines = torch.from_numpy(np.array(resampled_streamlines)) # Shape [N, n_vertices, 3]
    if verbose: print(f"Time to shorten streamlines: {time.time() - time_to_shorten_streamlines:.2f}s")
    return resampled_streamlines, np.array(origStreamlines, dtype=object)



@torch.inference_mode()
def shuffle_and_split_in_context_size(data_tensor: torch.Tensor, 
                                      origStreamlines: np.ndarray, 
                                      context_size: int,
                                      rng: torch.Generator):
    """
    Function that splits the streamlines into context_size parts. If the number of streamlines is not divisible by context_size,
    the function adds streamlines from the beginning of the tensor to the end.

    Args:
        data_tensor: The shortened streamlines.
        origStreamlines: The original streamlines.
        context_size: The size of the context.

    Returns:
        context_streamlines: The streamlines split into context_size parts. The length might differ from the original length.
        context_streamlines_original: The original streamlines split into context_size parts.
    """
    assert len(data_tensor) == len(origStreamlines), "The number of streamlines and the number of data tensor do not match."
    n_streamlines, nPoints, spaceDim = data_tensor.size()
    random_indices = torch.randperm(len(data_tensor), generator=rng)
    data_tensor = data_tensor[random_indices]
    if n_streamlines % context_size > 0:
        streamlines_to_add = context_size - n_streamlines % context_size
        data_tensor = torch.cat([data_tensor, data_tensor[: streamlines_to_add]], dim=0)
    origStreamlines = origStreamlines[random_indices] 
    return data_tensor.reshape(-1, context_size, nPoints, spaceDim), origStreamlines, random_indices



@torch.inference_mode()
def model_forward_pass(model: nn.Module, 
                       eval_batch_size: int,
                       data_tensor: torch.Tensor,
                       return_anatomical_clusters: bool,
                       mapping_800_800_to_43: torch.Tensor,
                       device: torch.device) -> torch.Tensor:
    """
    Forward pass of the data through the model. This function asumes that the set of streamlines is shuffled.
    Device transfer is done in an asynchronous fashion.  
    Args:
        model: The model to be evaluated.
        eval_batch_size: The batch size for the evaluation.
        device: The device on which the model is evaluated.
        data_tensor: The input data tensor (CPU). It is already normalized to the [-1,1] cube, Shape [N, context_size, n_vertices, 3]
    Returns:
        output_tensor: The output tensor (CPU) of the model.
    """
    non_blocking = False
    
    current_batch = data_tensor[0: eval_batch_size].to(device, non_blocking=non_blocking)
    model.eval()
    model = model.to(device, non_blocking=non_blocking)
    y_pred = []
    for i in tqdm(range(0, len(data_tensor), eval_batch_size)):
        next_batch = data_tensor[i + eval_batch_size : i + 2 * eval_batch_size].to(device, non_blocking=non_blocking)
        batch = current_batch # normalization_function(current_batch)
        output = model(batch) # Shape [bs, dim_out, context_size, dim_out]
        if return_anatomical_clusters:
            y_pred.append(mapping_800_800_to_43[output.argmax(dim=-1).flatten()].to("cpu", non_blocking=non_blocking)) # Shape [bs * context_size] in the range [0, 42]
        else:
            y_pred.append(output.argmax(dim=-1).flatten().to("cpu", non_blocking=non_blocking)) # Shape [bs * context_size] in the range [0, 1599]
        current_batch = next_batch
    
    # torch.cuda.synchronize(device)
    return torch.cat(y_pred, dim=0) # Shape [N x context_size]



def load_tcks(tck_path: Union[os.PathLike, str], verbose: bool) -> ArraySequence:
    """
    Function that reads a tck file.
    Args:
        tck_path: Path to the tck file.
        verbose: Whether timings should be printed 
    Returns:
        loaded_streamlines: streamlines loaded as an ArraySequence object
    """
    tck_path = Path(tck_path)
    assert tck_path.exists(), f"File {tck_path} does not exist."
    assert tck_path.suffix == ".tck", f"The file {tck_path} is not a '.tck' file."
    # Load Streamlines from tck file using nibabel
    time_to_load_tck = time.time()
    tractogram = nib.streamlines.load(tck_path)
    loaded_streamlines = tractogram.streamlines
    if verbose:
        print(f"Time to load tck file with {len(loaded_streamlines)} streamlines: {time.time() - time_to_load_tck:.2f}s")
    return loaded_streamlines



def results_to_tck(origStreamlines: Union[list, torch.Tensor, np.ndarray, ArraySequence], 
                   y_pred_43: torch.Tensor, 
                   int_to_label: np.ndarray, 
                   output_path: Union[os.PathLike, str]):
    """
    Function that writes the results to a tck file.
    Args:
        origStreamlines: A list of numpy arrays where each numpy array is a streamline from the tck file.
        y_pred_43: The predicted labels. Shape [N,] integer values in the range [0, 42].
        int_tpo_label: The mapping from integer to label. Each of the 43 labels can be accessed by the index.
    """
    assert len(origStreamlines) == len(y_pred_43), "The number of streamlines and the number of predictions do not match."
    # Create a new tractogram
    output_path = Path(output_path)
    y_pred = y_pred_43.numpy()
    if not output_path.exists():
        output_path.mkdir(parents=True)
    tcks = [tck_io.Tck(output_path / f"{label}.tck") for label in int_to_label]
    for tck in tcks:
        tck.force = True  # Force overwrite if the file already exists
    [tck.write({}) for tck in tcks]
    for k in range(len(int_to_label)):
        to_save = origStreamlines[y_pred == k]
        if len(to_save) > 0:
            to_save = np.ascontiguousarray(np.vstack([np.vstack([s, np.array([np.nan]*3)]) for s in to_save]), dtype='<f4')
            with open(output_path / f"{int_to_label[k]}.tck", "ab") as f:
                f.write(np.ndarray.tobytes(to_save))
    for tck in tcks:
        tck.close()


def print_command_line_bar_plot(y_pred: np.ndarray, int_to_label: np.ndarray):
    import math
    BLUE = "\033[34m"
    RESET = "\033[0m"
    MAX_LINE_WIDTH = 50

    counts, _ = np.histogram(y_pred, bins=43, range=(0, 42))
    max_number_digits = math.ceil(np.log(max(counts)+1) / np.log(10)) +1
    max_count = counts[:42].max()

    print(BLUE + "Class distribution" + RESET)
    print(BLUE + f"{"Class".center(2+15)} | Count" + RESET)

    for i, count in enumerate(counts[:42]):
        bar_len = math.ceil(count / max_count * MAX_LINE_WIDTH) 
        bar = "▇" * bar_len
        bin_label = int_to_label[i].ljust(15," ")
        print(BLUE + f"{str(i).rjust(2, " ")} {bin_label} {("("+str(count)).rjust(max_number_digits, " ") }) {bar}" + RESET)
    
    # Outlier Class
    bar_len = min(math.ceil(counts[-1] / max_count * MAX_LINE_WIDTH), MAX_LINE_WIDTH+1)
    bar = "#" * bar_len if math.ceil(counts[-1] / max_count * MAX_LINE_WIDTH) > MAX_LINE_WIDTH else "▇" * bar_len
    bin_label = "Outlier 'Other'".ljust(15," ")
    print(BLUE + f"{str(42).rjust(2, " ")} {bin_label} {("("+str(counts[-1])).rjust(max_number_digits, " ") }) {bar}" + RESET)


import yaml
from types import SimpleNamespace
from typing import Union
from pathlib import Path


def load_args_from_yaml(yaml_file_path: Union[str, Path]) -> SimpleNamespace:
    """ Little helper function to load the arguments from a yaml file """
    with open(yaml_file_path, 'r') as file:
        args = yaml.safe_load(file)
    return SimpleNamespace(**args)


import argparse
from .run import RapidParcTckEval
import torch

def main():
    parser = argparse.ArgumentParser(prog="rapidparc-run", description="Apply RapidParc on a .tck tractogram")
    parser.add_argument("--tck_path", type=str, required=True, help="Path to the input tck file.")
    parser.add_argument("--out_path", type=str, default="output_tcks", help="Path to the output folder, where the resulting tck files will be stored. If the folder does not exist, it will be created.")
    parser.add_argument("--model_name_or_path", type=str, required=True, help="One of the following: A pretrained model, i.e. one of 'rapidparc', 'rapidparc_v8' or 'hemiaug' Or a path of an experiment run (i.e. a folder of the `pretrained_models` folder after training)")
    parser.add_argument("--eval_batch_size", type=int, default=512, help="Batch size for evaluation.")
    parser.add_argument("--eval_context_size", type=int, default=2000, help="Context size for evaluation.")
    parser.add_argument("--device", type=str, default="cuda" if torch.cuda.is_available() else "cpu", help="Device to use for evaluation.")
    parser.add_argument("--seed", type=int, default=42, help="Random seed for reproducibility.")
    parser.add_argument("--print_time", action="store_true", help="Whether to print time measurements.")
    parser.add_argument("--print_class_distributiion", action="store_true", help="Whether the predicted class distribution should be printed")
    args = parser.parse_args()

    RapidParcTckEval(tck_path=args.tck_path,
                        out_path=args.out_path,
                        model_name_or_path=args.model_name_or_path,
                        eval_batch_size=args.eval_batch_size,
                        eval_context_size=args.eval_context_size,
                        seed=args.seed,
                        device=torch.device(args.device),
                        print_time=args.print_time,
                        print_class_distributiion=args.print_class_distributiion)



from typing import Union
from pathlib import Path
import tarfile
import pooch
import os
import numpy as np
import torch
from safetensors.torch import load_file

from .model_snapshot_handling import get_latest_snapshot
from .args_from_yaml import load_args_from_yaml
from .model import TransformerModel, get_embedding_layer

CACHE_DIR = Path(pooch.os_cache("RapidParc"))
MODEL_DIR = CACHE_DIR / "Models"
TRACTCLOUD_FOLDER = CACHE_DIR / "TrainData_800clu800ol"
PRETRAINED_MODELS = ["rapidparc", "rapidparc_v8", "hemiaug"]
BASE_DOWNLOAD_URL = "https://github.com/MedVisBonn/RapidParc/releases/download/v1.0.0"
HASHES = {
    "hemiaug.safetensors": "sha256:b66dfce6135aac56c3f835264e5ae9630b2497859425ad98cdc6d842c6f94435",
    "hemiaug_args.yaml": "sha256:4afe6fc28ee16f26ce98e529503a8cadaf792b5951a4293b1b89ad20c63c890c",
    "int_to_label.npy": "sha256:bbb0d6e8d2eb208677eded80ccc93f5d00bd1120e81e13fb5d5c6e20ba2650c0",
    "mapping_from_800_800_to_43.pt": "sha256:8836303db236fe885fbdcd827559cce69f4673b772fb6ae5d6caf78544fe9522",
    "rapidparc.safetensors": "sha256:6a158dad6a0b6124946da102b6037a6f7f8a16fb24e7caec15eaf95213e7b793",
    "rapidparc_args.yaml": "sha256:5c2b13a9843dc1fb73a066ccbfd0ce3455112048786f7e7ff594e6a4974bac4d",
    "rapidparc_v8.safetensors": "sha256:16a0e678b19a825177ae6bedcb068999befd7ba06163766f1b459b52a58b39e9",
    "rapidparc_v8_args.yaml": "sha256:1c54411e9eab69b0854f309ca47f6dfe05f125235df584664d28d2007cdf28dd"
}


def load_model_and_args_local_or_online(name_or_path: Union[str, os.PathLike], device: torch.device):
    if not isinstance(name_or_path, str) and not isinstance(name_or_path, os.PathLike):
        raise ValueError('Expected varaible name_or_path to be either one of "rapidparc", "rapidparc_v8" and "hemiaug"'\
                         +f'or to be an experiment path (containing model and args.yaml), but got {name_or_path}')
    if isinstance(name_or_path, str) and name_or_path in PRETRAINED_MODELS:
        ########## download args and model from Github ##########
        args_filename = f"{name_or_path}_args.yaml"
        pooch.retrieve(
            url=f"{BASE_DOWNLOAD_URL}/{args_filename}",
            known_hash=HASHES[args_filename],
            path=MODEL_DIR,
            fname=args_filename
        )
        
        args = load_args_from_yaml(MODEL_DIR / args_filename)

        ########## define model architecture ##########
        n_vertices = getattr(args, "n_vertices", None)
        if n_vertices is None:
            n_vertices = getattr(args, "num_support_points_per_streamline", None)
        if n_vertices is None:
            raise AttributeError(
                "Configuration is missing 'n_vertices'. Please add it to args.yml or provide "
                "'num_support_points_per_streamline'."
            )

        embedding = get_embedding_layer(
            num_support_points_per_streamline=n_vertices,
            d_model=args.d_model)

        model = TransformerModel(num_layers=args.num_layers, 
                            d_model=args.d_model,
                            nhead=args.nhead,
                            embedding_layer=embedding,
                            dim_feedforward=args.dim_feedforward,
                            dropout=args.dropout,
                            dim_class_hidden=args.dim_class_hidden,
                            dim_out=args.dim_out)
        
        ########## load weights of the model ##########
        model_filename = f"{name_or_path}.safetensors"
        pooch.retrieve(
            url=f"{BASE_DOWNLOAD_URL}/{model_filename}",
            known_hash=HASHES[model_filename],
            path=MODEL_DIR,
            fname=model_filename
        )

        state_dict = load_file(MODEL_DIR / model_filename) 
        model = model.to(device)  # deine Modellklasse
        model.load_state_dict(state_dict)
        model.eval()
        return model, args
    
    # Load Model from local folder:
    experiment_path = Path(name_or_path)
    assert experiment_path.is_dir(), f"The directiory {experiment_path} does not exist but should."
    
    if (experiment_path / "args" / "args.yaml").is_file():
        args = load_args_from_yaml(yaml_file_path=experiment_path / "args" / "args.yaml")
    elif (experiment_path / "args" / "args.yml").is_file():
        args = load_args_from_yaml(yaml_file_path=experiment_path / "args" / "args.yml")
    else:
        raise ValueError(f"File {experiment_path / "args" / "args.yaml"} does not exist but should.")
    
    n_vertices = getattr(args, "n_vertices", None)
    if n_vertices is None:
        n_vertices = getattr(args, "num_support_points_per_streamline", None)
    if n_vertices is None:
        raise AttributeError(
            "Configuration is missing 'n_vertices'. Please add it to args.yml or provide "
            "'num_support_points_per_streamline'."
        )
    
    embedding = get_embedding_layer(
        num_support_points_per_streamline=n_vertices,
        d_model=args.d_model)
    
    # Model
    model = TransformerModel(num_layers=args.num_layers, 
                             d_model=args.d_model,
                             nhead=args.nhead,
                             embedding_layer=embedding,
                             dim_feedforward=args.dim_feedforward,
                             dropout=args.dropout,
                             dim_class_hidden=args.dim_class_hidden,
                             dim_out=args.dim_out)
    model = model.to(device)
    assert (experiment_path/"model").is_dir(), f"The folder {experiment_path/"model"} does not exist but should"
    model_file = get_latest_snapshot(experiment_path/"model")
    if model_file is None:
        raise ValueError(f"Could not load the latest model. There is no model in {experiment_path/"model"}.")
    checkpoint = torch.load(model_file, weights_only=True, map_location=device)
    model = torch.nn.DataParallel(model)
    model.load_state_dict(checkpoint['model'])
    model = model.module
    model.eval()

    return model, args




def get_mapping_800_800_to_43_online() -> torch.Tensor:
    filename = "mapping_from_800_800_to_43.pt"
    pooch.retrieve(
        url=f"{BASE_DOWNLOAD_URL}/{filename}",
        known_hash=HASHES[filename],
        path=CACHE_DIR,
        fname=filename
    )
    return torch.load(CACHE_DIR/filename)


def get_int_to_label_online() -> np.ndarray:
    filename = "int_to_label.npy"
    pooch.retrieve(
        url=f"{BASE_DOWNLOAD_URL}/{filename}",
        known_hash=HASHES[filename],
        path=CACHE_DIR,
        fname=filename
    )
    return np.load(CACHE_DIR/filename)



def get_tractCloud_dataset_path(consent_given: bool = False) -> Path:
    # If the Dataset is downloaded, user already accepted licence agreement 
    if TRACTCLOUD_FOLDER.exists() and TRACTCLOUD_FOLDER.is_dir():
        return TRACTCLOUD_FOLDER

    print("\nFor running this code, you have to download the TractCloud dataset. \nPlease check out the Slicer Licence agreement before downloading the dataset:\n")
    print("\thttps://github.com/SlicerDMRI/TractCloud/blob/main/LICENSE\n")
    
    while not consent_given:
        answer = input("Do you accept the license agreement? [y/n]: ").strip().lower()
        if answer in ("y", "yes"):
            consent_given = True
        elif answer in ("n", "no"):
            raise PermissionError("Download aborted: License agreement not accepted.")
        else:
            print("Invalid input. Please answer with 'y' or 'n'.")

    # Run download:
    tar_url = "https://github.com/SlicerDMRI/TractCloud/releases/download/v1.0.0/TrainData_800clu800ol.tar.gz"
    xlsx_url = "https://github.com/SlicerDMRI/TractCloud/raw/2c2869cde12b8d617d21be8e19fd33ae73b5f3bc/datasets/FiberClusterAnnotation_Updated20230110.xlsx"

    tar_path = pooch.retrieve(
        url=tar_url, 
        known_hash="72d70d4cf7a465a824ff421f7df29ef9116ab2c890ad428d312f020704cdfac7", 
        path=CACHE_DIR
    )
    
    with tarfile.open(tar_path, "r:gz") as tar:
        tar.extractall(path=CACHE_DIR)
    
    pooch.retrieve(
        url=xlsx_url,
        known_hash="f4ff0162c5fe0a584d735f475260018bfdf52fdb9193b5303a8d24d02c1a6da9",
        path=TRACTCLOUD_FOLDER,  # Put annotation data in the same folder. 
        fname="FiberClusterAnnotation_Updated20230110.xlsx"
    )
    
    # Delete tar
    Path(tar_path).unlink(missing_ok=True)
    
    print(f"\nDownload complete - TractCloud Dataset is now stored at: {TRACTCLOUD_FOLDER}\n")
    return TRACTCLOUD_FOLDER




import torch
import torch.nn as nn
from pathlib import Path
from sklearn.metrics import f1_score, accuracy_score, classification_report
from typing import Optional, Union, Tuple
import numpy as np
import time
import os


from .utils.transforms3D import normalize_to_identity_cube, Rotate3DBatch_UnifDist
from .utils.dataset import get_TractCloud_eval_returner
from .utils.visualizations import create_plot
from .utils.pypi_package_helper import load_model_and_args_local_or_online, get_tractCloud_dataset_path, get_mapping_800_800_to_43_online, get_int_to_label_online


@torch.inference_mode()
def test(
        model_name_or_path: Union[str, os.PathLike], 
        applyTestSetAugmentations: bool,
        eval_batch_size: int = 128,
        eval_context_size: int = 2000,
        device: Optional[torch.device] = None, 
        augmentations: Optional[nn.Module] = None,
        print_classification_report: bool = True,
        tractcloud_licence_consent_given: bool = False
        ) -> Tuple[float, Union[float, np.ndarray]]:
    """ Benchmark function to test RapidParc on the tractcloud dataset test split. Creates Confusion matrices.

        Parameters:
            model_name_or_pathmodel_folder_path (PathLike | str):
                One of the following:
                A pretrained model, i.e. one of 'rapidparc', 'rapidparc_v8' or 'hemiaug'
                Or a path of an experiment run (i.e. a folder of the 'pretrained_models' folder after training)

            applyTestSetAugmentations (bool):
                Whether the 30 instances of the testset should be randomly rotated and appended to the test set. 

            eval_batch_size (int):
                Batch size during forward pass. One batch has the size of eval_batch_size * eval_context_size * n_verticies * 3
                Depending on your system, this value should be choosen to be small to prevent cuda-out-of-memory issues.

            eval_context_size (int):
                Context size during evaluation. Assumed to be a integral fraction of 10000. 
                Default is 2000. Larger values lead to slightly better results, but at the expense of higher computational effort.

            device (torch.device):
                Torch device used for evaluation. 'None' means: automatically chosen to be the best one available (if only one GPU is available).

            augmentations (torch.nn.Module):
                Optional augmentations used instead of the pre-defined random rotations. Only active if applyTestSetAugmentations is True.

            print_classification_report (bool):
                Whether the scipy classification report should be printed to the terminal.

            tractcloud_licence_consent_given (bool):
                Whether you agree to the slicer licence agreement (Only asked for once on each machine).
                This is the licence for using the TractCloud Dataset. For further information, take a look at
                https://github.com/SlicerDMRI/TractCloud/blob/main/LICENSE   

                
        Returns:
            Tuple of Accuracy and Macro F1 Score for y_true vs. y_pred.

    """
    
    if applyTestSetAugmentations:
        print(f"Testing experiment {model_name_or_path} on the test set with 30 augmentations")
    else: 
        print(f"Testing experiment {model_name_or_path} on the test set without augmentations")

    torch.set_float32_matmul_precision('high')
    if device is None:
        if torch.cuda.is_available(): device = torch.device("cuda")
        elif torch.backends.mps.is_available(): device = torch.device("mps")
        else: device = torch.device("cpu")
    map_800_800_to_43 = get_mapping_800_800_to_43_online().to(device)

    model, args = load_model_and_args_local_or_online(name_or_path=model_name_or_path, device=device)
    n_vertices = getattr(args, "n_vertices", None)
    
    if n_vertices is None:
        n_vertices = getattr(args, "num_support_points_per_streamline", None)
    if n_vertices is None:
        raise AttributeError(
            "Configuration is missing 'n_vertices'. Please add it to args.yml or provide "
            "'num_support_points_per_streamline'."
        )

    # Load test data
    subject_streamlines, subject_labels, _ = \
        get_TractCloud_eval_returner(
            inputPath=get_tractCloud_dataset_path(consent_given=tractcloud_licence_consent_given),
            device=torch.device("cpu"),
            eval_context_size=eval_context_size,
            n_vertices=n_vertices,
            split="test")()
    
    # Apply Augmentations
    if applyTestSetAugmentations:
        if augmentations is None:
            # Use TractCloud Augmentations (without translation and stretching because we normalize anyway)
            augmentations = nn.Sequential(
                Rotate3DBatch_UnifDist(range_x=[-45, 45], range_y=[-10, 10], range_z=[-10, 10])
            )# .to(device)
        # Copy data 30 times as TractCloud does
        streamlines = subject_streamlines.repeat_interleave(30, dim=0)
        labels_43 = subject_labels.repeat_interleave(30, dim=0) 
        streamlines = augmentations(streamlines)  
        # Add the original streamlines to the augmented ones (as TractCloud does)
        streamlines = torch.cat([subject_streamlines, streamlines], dim=0) # Added the original streamlines to the augmented ones
        labels_43 = torch.cat([subject_labels, labels_43], dim=0) # Shape [numSubjects * 31, numStreamlines]
    else:
        # Nothing is done here
        streamlines = subject_streamlines 
        labels_43 = subject_labels 
    
    # Shuffle Streamlines for each Subject
    streamlines, labels_43 = torch.vmap(random_permutate_streamlines, randomness="different")(streamlines, labels_43)
    _, numStreamlinesPerSubject, seqLength, spaceDim = streamlines.shape
    assert numStreamlinesPerSubject % eval_context_size == 0, \
        f"In this script, we assume that numStreamlinesPerSubject % eval_context_size == 0, "\
        +"but got numStreamlinesPerSubject={numStreamlinesPerSubject} and eval_context_size={eval_context_size}"
    streamlines = streamlines.cpu()
    labels_43 = labels_43.flatten().cpu().numpy()
    
    # Start testing and timing
    duration = time.time()
    streamlines = streamlines.to(device)
    map_800_800_to_43 = map_800_800_to_43.to(device)
    streamlines = streamlines.reshape(-1, eval_context_size, seqLength, spaceDim)
    streamlines = normalize_to_identity_cube(streamlines)

    y_pred_800_800 = []
    for i in range(0, len(streamlines), eval_batch_size):
        batch = streamlines[i : i + eval_batch_size]
        y_pred_800_800_batch = model(batch).argmax(dim=-1).flatten() # Shape [batch_size * context_size]
        y_pred_800_800.append(y_pred_800_800_batch)
    
    y_pred_800_800 = torch.cat(y_pred_800_800, dim=0) # Shape [numSubjects * 31 * context_size]
    y_pred = map_800_800_to_43[y_pred_800_800].cpu().numpy()

    duration = time.time() - duration
    print(f"Duration: {duration:.2f} seconds")
    target_names = get_int_to_label_online()

    experiment_path = Path(model_name_or_path)
    if not experiment_path.is_dir():
        experiment_path.mkdir()

    if print_classification_report:
        print(classification_report(y_true=labels_43, y_pred=y_pred, digits=5, target_names=target_names, zero_division=0))
    
    if applyTestSetAugmentations:
        save_dir = Path(experiment_path / "plots" / "testset_STA")
    else:
        save_dir = Path(experiment_path / "plots" / "testset")
    if not save_dir.exists():
        save_dir.mkdir(parents=True)
    create_plot(y_true=labels_43, 
                y_pred=y_pred, 
                subject_number=None, 
                duration=duration, 
                save_path=save_dir,
                classnames=target_names, 
                unnormalized=False)
    return accuracy_score(y_true=labels_43, y_pred=y_pred), f1_score(y_true=labels_43, y_pred=y_pred, average="macro")



@torch.inference_mode()
def random_permutate_streamlines(streamlines, labels):
    assert len(streamlines.shape) == 3, f"Streamlines have wrong shape: {streamlines.shape}, but assumed something like [numStreamlines, 15, 3]"
    indices = torch.randperm(streamlines.shape[0])
    return streamlines[indices], labels[indices]



import os 
import sys
import yaml
from pathlib import Path
import torch
import torch.nn as nn
import torch.distributed as dist
from torch.nn.parallel import DistributedDataParallel as DDP
from torch.distributed.elastic.multiprocessing.errors import record
from typing import Union

from .utils.dataset import get_TractCloud_train_returner, get_TractCloud_eval_returner, get_mapping_from_800_800_to_43, get_int_to_label
from .utils.model import TransformerModel, get_embedding_layer
from .utils.trainer import train as _train
from .utils.visualizations import create_confusion_matrix
from .utils.transforms3D import RandomNoiseBatch, Rotate3DBatch_UnifDist, RandomNoisedNormalizeToIdentetyCube
from .utils.ddp_handling import ddp_setup, ddp_cleanup, ddp_is_running
from .utils.pypi_package_helper import get_tractCloud_dataset_path
from .test import test



def create__output_folders(output_folder_name: Union[str, os.PathLike], subfolders: list = ["model", "args", "slurm", "plots"]):
    output_folder_name = Path(output_folder_name)
    if not output_folder_name.is_dir(): 
        output_folder_name.mkdir()
    for subfolder in subfolders:
        if not (output_folder_name / subfolder).is_dir(): 
            (output_folder_name / subfolder).mkdir()


@record
def train(args) -> None: 
    """
    Training function to retrain RapidParc.
    
    Parameters:
        - args data object:
            For definition of the args-Object, take a look at `cli_train.py`
            If you want to call this function from python, you may define args as a dataclass instance.
    """

    torch.set_float32_matmul_precision('high')
    if args.use_ddp:   
        ddp_setup()
        args.local_batch_size = args.global_batch_size // dist.get_world_size() 
        device = torch.device(type="cuda", index=int(os.environ['LOCAL_RANK']))
    else: 
        args.local_batch_size = args.global_batch_size
        if torch.cuda.is_available():
            device = torch.device("cuda")
        elif torch.backends.mps.is_available():
            device = torch.device("mps")
        else:
            device = torch.device("cpu")

    os.chdir(str(Path(__file__).parent.resolve()))

    # Create output folders and wandb init
    output_folder = Path(args.out_dir)
    if not args.use_ddp or dist.get_rank() == 0:
        if not output_folder.is_dir():
            output_folder.mkdir()
        create__output_folders(output_folder_name = output_folder / args.slurm_job_id)
    output_folder = output_folder / args.slurm_job_id

    # Load Dataloaders
    data_path = get_tractCloud_dataset_path(consent_given=args.tractcloud_licence_consent_given) 
    get_train_Tensor = get_TractCloud_train_returner(
        inputPath=data_path,
        device=device,
        n_vertices=args.n_vertices)
    get_val_Tensor = get_TractCloud_eval_returner(
        inputPath=data_path,
        device=device,
        eval_context_size=args.context_size,
        n_vertices=args.n_vertices,
        split="val")
    
    map_800_800_to_43 = torch.from_numpy(get_mapping_from_800_800_to_43(data_path))
    
    # Augmentions (on GPU)
    trainAugmentations = nn.Sequential(
        Rotate3DBatch_UnifDist(range_x=args.aug_rotate_unif_x, range_y=args.aug_rotate_unif_y, range_z=args.aug_rotate_unif_z),
        RandomNoiseBatch(mu=0, sigma=args.aug_noise_sigma),
        RandomNoisedNormalizeToIdentetyCube(shift_std=args.aug_norm_shift_std, scale_std=args.aug_norm_scale_std)
    ).to(device)
    
    embedding= get_embedding_layer(
        num_support_points_per_streamline=args.n_vertices,
        d_model=args.d_model)

    # Model
    model = TransformerModel(num_layers=args.num_layers, 
                             d_model=args.d_model,
                             nhead=args.nhead,
                             embedding_layer=embedding,
                             dim_feedforward=args.dim_feedforward,
                             dropout=args.dropout,
                             dim_class_hidden=args.dim_class_hidden,
                             dim_out=args.dim_out)
    model.to(device)

    if not args.use_ddp or dist.get_rank() == 0:
        print(f"Number Devices available: {torch.cuda.device_count()}")
        print("############### MODEL ###############")
        print(model)
        print(f"Number trainable parameters: {sum(p.numel() for p in model.parameters() if p.requires_grad)} \n")
    if args.use_ddp:
        print(f"Rank {dist.get_rank()} is online and handles GPU {device}")
        model = DDP(model, device_ids=[device], find_unused_parameters=True)
    else:
        model = nn.DataParallel(model)


    criterion = nn.CrossEntropyLoss()
    optimizer = torch.optim.AdamW(model.parameters(), lr=args.lr, weight_decay=0.001)

    if args.sched_args["scheduler"] == "CosineAnnealingWarmRestarts":
        # If this scheduler is used, I recommend to add gradient clipping
        scheduler = torch.optim.lr_scheduler.CosineAnnealingWarmRestarts(optimizer, T_0=args.sched_args["T_0"], T_mult=args.sched_args["T_mult"], eta_min=args.sched_args["eta_min"])
    elif args.sched_args["scheduler"] == "CosineAnnealingLR":
        scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(optimizer, T_max=args.num_epochs, eta_min=args.sched_args["eta_min"])
    else:
        raise NotImplementedError(f"Scheduler {args.sched_args['scheduler']} not implemented")
    
    # Store args
    if not ddp_is_running() or dist.get_rank() == 0:
        with open(str(Path(output_folder) / 'args' / 'args.yaml'), 'w') as outfile:
            yaml.dump(vars(args), outfile, default_flow_style=False, sort_keys=False)

    # Training
    model = _train(
        get_train_Tensor=get_train_Tensor,
        get_val_Tensor=get_val_Tensor,
        num_epochs=args.num_epochs,
        batch_size=args.local_batch_size,
        context_size=args.context_size,
        optimizer=optimizer,
        scheduler=scheduler,
        model=model,
        device=device,
        criterion=criterion,
        snapshot_path=str(Path(output_folder) / "model"),
        snapshot_every=args.snapshot_every,
        store_all_snapshots=args.store_all_snapshots,
        transforms_on_gpu=trainAugmentations,
        aug_p_hemi = args.aug_p_hemi,
        aug_method_hemi = args.aug_method_hemi,
        map_800_800_to_43=map_800_800_to_43
        )
    
    # Confusion Matrix and dumping args
    if not ddp_is_running() or dist.get_rank() == 0: 
        sys.stdout.flush()
        print("\n\n##### Confusion Matricies #####")

        for split in ["train", "val", "test"]:
            print(f"\nCreating confusion matricies on {split} set")
            create_confusion_matrix(
                model=model,
                split=split,
                input_path=data_path,
                save_path=str(Path(output_folder) / "plots" / split),
                device=device,
                map_800_800_to_43=map_800_800_to_43,
                int_to_label=get_int_to_label(data_path),
                n_vertices = args.n_vertices,
                eval_batch_size=64
            )

        
        accuracy_STA, macro_f1_score_STA = test(
            model_name_or_path = output_folder, 
            applyTestSetAugmentations=True,
            eval_batch_size=8, 
            eval_context_size=2000, 
            device=device,
            print_classification_report=False)
        
        accuracy, macro_f1_score = test(
            model_name_or_path = output_folder, 
            applyTestSetAugmentations=False,
            eval_batch_size=8, 
            eval_context_size=2000, 
            device=device,
            print_classification_report=False)

        print("Final results on TractCloud test set\n" 
              +f"- Acc: {100*accuracy:.2f}%, macro F1: {100*macro_f1_score:.2f}% \n"\
              +f"- STA Acc: {100*accuracy_STA:.2f}%, STA Macro F1: {100*macro_f1_score_STA:.2f}%")
    if ddp_is_running():
        print(f"Max memory used by device {device}: {torch.cuda.max_memory_allocated(device=device) / 1024**3}")
        sys.stdout.flush()
        dist.barrier()
        ddp_cleanup()


import torch
import torch.nn as nn
from torch import distributed as dist
from typing import Tuple


from .transforms3D import normalize_to_identity_cube, RandomFlipBatch
from .ddp_handling import ddp_is_running



class TransformerModel(nn.Module):
    def __init__(self, dim_class_hidden: int, dim_out: int, num_layers: int, embedding_layer,  **transformer_args):
        super(TransformerModel, self).__init__()
        self.d_model = transformer_args['d_model']
        self.dim_class_hidden = dim_class_hidden
        self.dim_out = dim_out
        self.embedding_layer = embedding_layer
        self.transformerEncoderLayer = nn.TransformerEncoderLayer(**transformer_args, batch_first=True)
        self.embeddingTransformer = nn.TransformerEncoder(encoder_layer = self.transformerEncoderLayer, 
                                                          num_layers = num_layers,
                                                          enable_nested_tensor = False)
        self.classifier = nn.Sequential(
            nn.Dropout(transformer_args["dropout"]),
            nn.Linear(self.d_model, self.dim_class_hidden),
            nn.ReLU(),
            nn.Linear(self.dim_class_hidden, self.dim_out)
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # Input-shape: [bs, context_size, 15, 3]
        x = normalize_to_identity_cube(x)
        x = self.embedding_layer(x) # [bs, context_size, d_model] ≈ [bs, 2000, 128]
        x = self.embeddingTransformer(x) # Shape [bs, context_size, d_model]
        x = self.classifier(x) # Shape [bs, context_size, dim_out]
        return x



def get_embedding_layer(num_support_points_per_streamline: int, d_model: int = 128) -> nn.Module:
    assert d_model >= 45, "Embedding type flip_aug only works with d_model>=45"
    return EmbeddingAugFlipBatchGPU(flipProbability=0.5, 
                                    num_support_points_per_streamline=num_support_points_per_streamline, 
                                    d_model=d_model)



class EmbeddingAugFlipBatchGPU(nn.Module):
    """
    Embed the input tensor with a flip augmentation based embedding. It fills the rest of the embedding with a linear layer.
    This class is designed to be used with the GPU with batch processing.
    """
    def __init__(self, flipProbability: float = 0.5, num_support_points_per_streamline: int = 15, d_model: int = 66):
        super(EmbeddingAugFlipBatchGPU, self).__init__()
        self.flipAugEmbedding = RandomFlipBatch(flipProbability=flipProbability)
        if d_model > 3 * num_support_points_per_streamline:
            self.adaptive = True
            self.linear = nn.Linear(in_features = num_support_points_per_streamline * 3, 
                                    out_features = d_model - num_support_points_per_streamline * 3)
            self.activation = nn.LeakyReLU()
        elif d_model == 3 * num_support_points_per_streamline:
            # Module reduces to flip augmentation and reshaping
            self.adaptive = False
        else:
            raise ValueError(f"d_model = {d_model} must be greater or equal" \
                             +"to 3 * numStreamlines = {3*num_support_points_per_streamline}")
        
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Parameters:
            x (torch.Tensor): Shape [bs, numStreamlines, numPoints, 3], on device

        Returns:
            torch.Tensor: Embedded Streamlines with shape [bs, numStreamlines, d_model]
        """
        bs, numStreamlines, _, _ = x.size()
        if self.training: # Only apply flip augmentation during training (model.train())
            x = self.flipAugEmbedding(x) # Shape [bs, numStreamlines, numPoints, 3]
        x = x.reshape(bs, numStreamlines, -1) # Shape [bs, numStreamlines, numPoints * 3]
        if self.adaptive:
            rest_embedding = self.activation(self.linear(x)) # Shape [bs, numStreamlines, d_model - numPoints * 3]
            x = torch.cat([x, rest_embedding], dim=2)
        return x # Shape [bs, numStreamlines, d_model]
    
    def __repr__(self):
        if self.adaptive:
            return f"{self.__class__.__name__}" + f"flipAugEmbedding={self.flipAugEmbedding}, linear={self.linear}, activation={self.activation}"
        else:
            return f"{self.__class__.__name__}" + f"flipAugEmbedding={self.flipAugEmbedding}"


import torch
import torch.nn as nn
from typing import List
import math


class Rotate3DBatch_UnifDist(nn.Module):
    """
    Rotate the input 3D tensor around the x-, y- and z-axis. This class is designed to be used with the GPU with batch processing.
    """

    def __init__(self, range_x: List[float], range_y: List[float], range_z: List[float]):
        """
        Parameters:
            range_x (List[float]): Range of the uniform distribution for the rotation around the x-axis in degrees
            range_y (List[float]): Range of the uniform distribution for the rotation around the y-axis in degrees
            range_z (List[float]): Range of the uniform distribution for the rotation around the z-axis in degrees
        """
        super(Rotate3DBatch_UnifDist, self).__init__()
        assert len(range_x) == 2, "range_x must have length 2"
        assert len(range_y) == 2, "range_y must have length 2"
        assert len(range_z) == 2, "range_z must have length 2"
        self.range_x = torch.tensor([range_x[0], range_x[1]]) * math.pi / 180
        self.range_y = torch.tensor([range_y[0], range_y[1]]) * math.pi / 180
        self.range_z = torch.tensor([range_z[0], range_z[1]]) * math.pi / 180


    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Parameters:
            x (torch.tensor): Shape [bs, numStreamlines, numPoints, 3], on device
        """
        bs = x.size(0)
        device = x.device
        thetaX = torch.rand(size=(bs,), device=device) * (self.range_x[1] - self.range_x[0]) + self.range_x[0] # Shape [bs]
        thetaY = torch.rand(size=(bs,), device=device) * (self.range_y[1] - self.range_y[0]) + self.range_y[0] # Shape [bs]
        thetaZ = torch.rand(size=(bs,), device=device) * (self.range_z[1] - self.range_z[0]) + self.range_z[0] # Shape [bs]

        rotationMatrix = \
            torch.stack([
                torch.stack([
                    torch.cos(thetaX) * torch.cos(thetaY), 
                    torch.cos(thetaX) * torch.sin(thetaY) * torch.sin(thetaZ) - torch.sin(thetaX) * torch.cos(thetaZ), 
                    torch.cos(thetaX) * torch.sin(thetaY) * torch.cos(thetaZ) + torch.sin(thetaX) * torch.sin(thetaZ)], dim=1), # Shape [bs, 3]
                torch.stack([
                    torch.sin(thetaX) * torch.cos(thetaY), 
                    torch.sin(thetaX) * torch.sin(thetaY) * torch.sin(thetaZ) + torch.cos(thetaX) * torch.cos(thetaZ), 
                    torch.sin(thetaX) * torch.sin(thetaY) * torch.cos(thetaZ) - torch.cos(thetaX) * torch.sin(thetaZ)], dim=1),
                torch.stack([
                    -torch.sin(thetaY), 
                    torch.cos(thetaY) * torch.sin(thetaZ), 
                    torch.cos(thetaY) * torch.cos(thetaZ)], dim=1)], 
            dim=2) # Shape [bs, 3, 3]
        del thetaX, thetaY, thetaZ

        return torch.einsum("bij, bspj -> bspi", rotationMatrix, x)
    

    def __repr__(self):
        return f"{self.__class__.__name__}: range_x={self.range_x}, range_y={self.range_y}, range_z={self.range_z}"



class RandomNoiseBatch(nn.Module):
    """
    Add random noise to the input tensor. This class is designed
    to be used with the GPU with batch processing.
    """
    def __init__(self, mu: float, sigma: float):
        """ 
        Parameters:
            mu (float): Mean of the normal distribution
            sigma (float): Standard deviation of the normal distribution
        """
        super(RandomNoiseBatch, self).__init__()
        self.mu = mu
        self.sigma = sigma

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Parameters:
            x (torch.tensor): Shape [bs, numStreamlines, numPoints, 3], on device
        """
        device = x.device
        noise = torch.normal(mean=self.mu, std=self.sigma, size=x.size(), device=device)
        return x + noise
    
    def __repr__(self):
        return f"{self.__class__.__name__}: mu={self.mu}, sigma={self.sigma}"
    


class RandomStretchBatch(nn.Module):
    """
    Stretch the input 3D tensor along the x-, y- and z-axis. All Streamlines of one subject are stretched in the same way.
    This class is designed to be used on GPU with batch processing.
    """
    def __init__(self, sigma_x: float, sigma_y: float, sigma_z: float):
        """ 
        Parameters:
            sigma_x (float): Standard deviation of the normal distribution along x-axis
            sigma_y (float): Standard deviation of the normal distribution along y-axis
            sigma_z (float): Standard deviation of the normal distribution along z-axis
        
        Output:
            torch.tensor: Stretched tensor with shape [bs, numStreamlines, numPoints, 3]
        """
        super(RandomStretchBatch, self).__init__()
        self.sigmas = torch.tensor([sigma_x, sigma_y, sigma_z], requires_grad=False)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Parameters:
            x (torch.tensor): Shape [bs, numStreamlines, numPoints, 3], on device
        """
        device = x.device
        bs = x.size(0)
        stretch = torch.normal(mean=1.0, std=self.sigmas.repeat(bs, 1)).to(device) # [bs, 3]
        return torch.einsum('bjkd,bd->bjkd', x, stretch)
    
    def __repr__(self):
        return f"{self.__class__.__name__}: sigma_x={self.sigmas[0]}, sigma_y={self.sigmas[1]}, sigma_z={self.sigmas[2]}"



class RandomFlipBatch(nn.Module):
    """
    Flip the input tensor along the Axis of the streamline supporting points. 
    This class is designed to be used with the GPU with batch processing.
    """
    def __init__(self, flipProbability: float = 0.5):
        super(RandomFlipBatch, self).__init__()
        self.flipProbability = flipProbability

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Parameters:
            x (torch.tensor): Shape [bs, numStreamlines, numPoints, 3], on device
        """
        device = x.device
        bs, numStreamlines, numPoints, dim = x.size()
       
        # Decide for each streamline of a batch if it should be flipped
        isFlipPosition = torch.bernoulli(self.flipProbability * torch.ones(size=(bs, numStreamlines), device=device)) # Shape [bs, numStreamlines]
        
        # Flip the streamlines if the flipPosition is 1
        return torch.einsum("bspd, bs -> bspd", torch.flip(x, dims=[2]), isFlipPosition) + torch.einsum("bspd, bs -> bspd", x, 1 - isFlipPosition)
    
    def __repr__(self):
        return f"{self.__class__.__name__}"



def normalize_to_identity_cube(x: torch.Tensor) -> torch.Tensor:
    """
    Normalize the input streamlines. This function is designed to be used with the GPU.
    The normalized streamlines should have a minimal value of -1 and a maximal value of 1 in all axis.

    Parameters:
        x (torch.tensor): Shape [bs, numStreamlines, numPoints, 3]

    Returns:
        torch.tensor: Normalized Streamlines with shape [bs, numStreamlines, numPoints, 3]
    """
    bs, _, _, dim = x.size()
    mins = torch.amin(x, dim=(1,2)).view(bs, 1, 1, dim)
    maxs = torch.amax(x, dim=(1,2)).view(bs, 1, 1, dim)

    return  2 * (x - mins) / (maxs - mins) -1.0



class RandomNoisedNormalizeToIdentetyCube(nn.Module):
    """
    Normalize the input streamlines. But it applies some random noise. This class is designed to be used with the GPU.
    The normalized streamlines should have a minimal value of -1 and a maximal value of 1 in all axis.
    """
    def __init__(self, shift_std: float = 0.01, scale_std: float = 0.01):
        """ 
        Parameters:
            shift_std (float): Standard deviation of the normal distribution for the shift
            scale_std (float): Standard deviation of the normal distribution for the scale
        """
        super(RandomNoisedNormalizeToIdentetyCube, self).__init__()
        self.shift_std = shift_std
        self.scale_std = scale_std

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Parameters:
            x (torch.tensor): Shape [bs, numStreamlines, numPoints, 3], on device
        """
        bs, _, _, dim = x.size()
        mins = torch.amin(x, dim=(1,2)).view(bs, 1, 1, dim)
        maxs = torch.amax(x, dim=(1,2)).view(bs, 1, 1, dim)

        shift_noise = torch.normal(mean=-1.0, std=self.shift_std, size=(bs, dim), device=x.device).view(bs, 1, 1, dim)
        scale_noise = torch.normal(mean=2.0, std=self.scale_std, size=(bs, dim), device=x.device).view(bs, 1, 1, dim)

        return  scale_noise * (x - mins) / (maxs - mins) + shift_noise
    
    def __repr__(self):
        return f"{self.__class__.__name__}: shift_std={self.shift_std}, scale_std={self.scale_std}"


import torch
from pathlib import Path
import numpy as np
import pickle
import pandas as pd
from collections import defaultdict
from typing import Tuple, Optional, Union, Dict, List
from collections.abc import Callable
import os



def pre_normalization(x: torch.Tensor) -> torch.Tensor:
    """
    Prenormalization used for TractCloud Dataset during training
    Parameters:
        x (torch.tensor): Shape [bs, numStreamlines, numPoints, 3]
    
    Returns:
        torch.tensor: Normalized Streamlines with shape [bs, numStreamlines, numPoints, 3]
    """
    bs, _, _, dim = x.size()

    # move minimal streamline coordinate to (0,0,0)
    x -=  torch.amin(x, dim=(1,2)).view(bs, 1, 1, dim)
    
    # scale s.t. maximal value is 2
    x /= 0.5 * torch.amax(x)
    
    # transform middle point of each axis to (0,0,0)
    x -= 0.5 * torch.amax(x, dim=(1,2)).view(bs, 1, 1, dim)
    
    assert torch.allclose(-torch.amin(x, dim=(1,2)), torch.amax(x, dim=(1,2))), "The Result of normalization is not centered around (0,0,0)"
    return x



def get_TractCloud_train_returner(inputPath: Union[str, os.PathLike],
                 device: torch.device,
                 n_vertices: Optional[int] = None) \
                    -> Callable[[Union[List[int], torch.Tensor]], Tuple[torch.Tensor, torch.Tensor]]:
    """
        This functions is an alternative to Pytorch Datasets and Dataloaders.
        In our case, transferring the data to the GPU took a relevant amount of time.
        Therefore, this closure leaves the data permanently on the GPU. 
        It returns two functions: one for the training batch and one for the validation set. 

        Parameters:
            inputPath:
                Path to the TractCloud Dataset Folder

            device:
                The device for the train set

            n_vertices (optional, int): 
                The number of verticies requested for further downsampling

        Returns:
            train_returner ((idx) -> streamlines[idx], labels_800_800[idx]):
                A function that returns the tractogram of a subject with this idx
    """

    inputPath = Path(inputPath)
    assert inputPath.exists(), f"Path {inputPath} does not exist"
    streamlines, _, labels_800_800 = getTractCloudDataset(tractCloud_Dataset_dir=inputPath, split= "train")

    # Convert to torch tensors and move to device
    streamlines = torch.from_numpy(streamlines).to(device, non_blocking=True) # Shape [numSubjects, numStreamlines, seqLength, spaceDim]
    labels_800_800 = torch.from_numpy(labels_800_800).to(device, non_blocking=True) # Shape [numSubjects, numStreamlines]
    streamlines = pre_normalization(streamlines) # Shape [numSubjects, numStreamlines, seqLength, spaceDim]
    if n_vertices is not None:
        streamlines = _downsample_streamlines(streamlines, n_vertices=n_vertices) # Shape [numSubjects, numStreamlines, n_vertices, spaceDim]

    def train_returner(indices: Union[List[int], torch.Tensor]) -> Tuple[torch.Tensor, torch.Tensor]:
        # the streamlines are normalized at that point
        return streamlines[indices], labels_800_800[indices] 
    
    return train_returner




def get_TractCloud_eval_returner(inputPath: Union[str, os.PathLike],
                 device: torch.device,
                 eval_context_size: int,
                 seed: int = 42,
                 n_vertices: Optional[int] = None,
                 split: str = "val") \
                    -> Callable[[], Tuple[torch.Tensor, torch.Tensor, torch.Tensor]]:
    """
        This functions is an alternative to Pytorch Datasets and Dataloaders.
        In our case, transferring the data to the GPU took a relevant amount of time.
        Therefore, this closure leaves the data permanently on the GPU. 
        It returns two functions: one for the training batch and one for the validation set. 

        Parameters:
            inputPath:
                Path to the TractCloud Dataset Folder

            device:
                The device for the train set
            
            eval_context_size:
                The constext size used for evaluation

            seed (int, default 42):
                The random seed used for permuting the validation tractograms.

            n_vertices (optional, int): 
                The number of verticies requested for further downsampling

            split (str):
                Defines what split should be returned, in ["train", "val", "test"]

        Returns:
            eval_returner (() -> streamlines, labels_43, labels_800_800):
                A function that returns the whole validation set with shape [bs, eval_context_size, seqLength, 3]
    """
    assert split in ["train", "val", "test"], f'split = {split} must be in ["train", "val", "test"]'
    inputPath = Path(inputPath)
    assert inputPath.exists(), f"Path {inputPath} does not exist"
    streamlines, labels_43, labels_800_800 = getTractCloudDataset(tractCloud_Dataset_dir=inputPath, split=split)

    rng = np.random.default_rng(seed)
    
    # Permute the streamlines for random subtractogram partitioning
    for i in range(len(streamlines)):
        random_permutation = rng.permutation(streamlines[i].shape[0])
        streamlines[i] = streamlines[i][random_permutation]
        labels_43[i] = labels_43[i][random_permutation]
        labels_800_800[i] = labels_800_800[i][random_permutation]

    # Convert to torch tensors
    streamlines = torch.from_numpy(streamlines).to(device) # Shape [numSubjects, numStreamlines, seqLength, spaceDim]
    labels_43 = torch.from_numpy(labels_43).to(device) # Shape [numSubjects, numStreamlines]
    labels_800_800 = torch.from_numpy(labels_800_800).to(device) # Shape [numSubjects, numStreamlines]

    # Reshape the streamlines to context size bits
    _, numStreamlinesPerSubject, seqLength, spaceDim = streamlines.shape
    assert numStreamlinesPerSubject % eval_context_size == 0, \
        f"For faster evaluation, we assume that numStreamlinesPerSubject % context_size == 0, " +\
        f"but got numStreamlinesPerSubject={numStreamlinesPerSubject} and context_size={eval_context_size}"
    streamlines = streamlines.reshape(-1, eval_context_size, seqLength, spaceDim)
    streamlines = pre_normalization(streamlines)
    labels_43 = labels_43.reshape(-1, eval_context_size)
    labels_800_800 = labels_800_800.reshape(-1, eval_context_size)
    
    if n_vertices is not None:
        # Downsample streamlines to n_vertices preserving both endpoints!
        streamlines = _downsample_streamlines(streamlines, n_vertices)

    def eval_returner() -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        # they are normed at that point
        return streamlines, labels_43, labels_800_800
    
    return eval_returner



def getTractCloudDataset(tractCloud_Dataset_dir: Union[str, os.PathLike], split: str) \
    -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
        This function returns the TractCloud dataset as reshaped Numpy Arrays and two helper arrays
        Parameters:
            tractCloud_Dataset_dir: 
                Path of the TractCloud TrainData_800clu800ol folder

            split (str):
                Defines what split should be returned, in ["train", "val", "test"]
        
        Returns:
            streamlines (np.ndarray, shape [N, 10000, 15, 3]): 
                10000 streamlines of N subjects of length 15 each 

            labels_43 (np.ndarray, shape [N, 10000]):
                The labels of the streamlines for the TractCloud dataset, each value in 0, ... 42 
                
            labels_800_800 (np.ndarray, shape [N, 10000]): 
                The labels of the streamlines in the org atlas, each value in 0, ..., 1599
    """
    assert split in ["train", "val", "test"], f'Split must be in ["train", "val", "test"] but got {split}'

    input_path = Path(tractCloud_Dataset_dir)
    assert input_path.is_dir(), f"Expected {input_path} to be the path of the unzipped TractCloud Train dataset"

    with open(input_path/f'{split}.pickle', 'rb') as f:
        data = pickle.load(f)
        
    streamlines = data["feat"].reshape(-1, 10000, 15, 3)
    labels_800_800 = data["label"].reshape(-1, 10000)
    labels_43 = get_mapping_from_800_800_to_43(input_path=input_path)[labels_800_800]

    # # Possible check if reshaping is correctly
    # subjectIDs = data["subject_id"].reshape(-1, 10000)
    # flat_stls = data["feat"]
    # flat_subIDS = data["subject_id"]
    # flat_labels = data["label"]
    # iFlat, iRsh, jRsh = 0, 0, 0
    # while iFlat < len(flat_stls):
    #     assert (streamlines[iRsh, jRsh] == flat_stls[iFlat]).all()
    #     assert (subjectIDs[iRsh, jRsh] == flat_subIDS[iFlat]).all()
    #     assert (labels_800_800[iRsh, jRsh] == flat_labels[iFlat]).all()
    #     iFlat += 1
    #     jRsh += 1
    #     if jRsh >= 10000:
    #         jRsh = 0
    #         iRsh += 1

    return streamlines, labels_43, labels_800_800



def get_int_to_label(input_path: Union[str, os.PathLike], save_path: Optional[Union[str, os.PathLike]] = None) -> np.ndarray:
    """
    Getter function for array that maps all tract indices 0, 1, ..., 42 
    to their names, i.e. "AF", "CB", ..., "Other"

    Parameters:
        input_path:
            Input folder path of tractCloud training data

    Returns:
        int_to_label (np.ndarray, shape [43], str):
            Represents a mapping from the 43 classes to their names
    """
    input_path = Path(input_path)
    int_to_label = np.array(list(obtain_TractClusterMapping(input_path).keys()))
    if save_path:
        save_path = Path(save_path) / "int_to_label.npy"
        if not save_path.exists():
            np.save(arr=int_to_label, file=save_path)
    return int_to_label



def get_mapping_from_800_800_to_43(input_path: Union[str, os.PathLike], save_path: Optional[Union[str, os.PathLike]] = None) -> np.ndarray:
    input_path = Path(input_path)
    mapping_from_800_800_to_43 = np.array(
        cluster2tract_label(
            lst = [i for i in range(1600)], 
            ordered_tract_cluster_mapping_dict=obtain_TractClusterMapping(input_path), 
            output_lst=False)
        )
    if save_path:
        save_path = Path(save_path) / "mapping_from_800_800_to_43.npy"
        if not save_path.exists():
            np.save(arr=mapping_from_800_800_to_43, file=save_path)
    return mapping_from_800_800_to_43



def _downsample_streamlines(streamlines: torch.Tensor, n_vertices: int) -> torch.Tensor:
    # Downsample streamlines to n_vertices preserving both endpoints!
    num_subjects, num_streamlines, seqLength, spaceDim = streamlines.shape
    if seqLength <= n_vertices:
        return streamlines  # No downsampling needed
    indices = torch.linspace(0, seqLength - 1, steps=n_vertices).long().to(streamlines.device)
    return streamlines[:, :, indices, :]




# ********************* Functions that originate from the TractCloud repository *********************

def cluster2tract_label(lst: Union[List, np.ndarray], 
                        ordered_tract_cluster_mapping_dict: Dict[str, List[str]], 
                        output_lst: bool = True
                        ) -> Union[List, np.ndarray]:
    """
        Convert cluster labels to tract labels

        This function is copied from TractCloud
        https://github.com/SlicerDMRI/TractCloud/blob/2c2869cde12b8d617d21be8e19fd33ae73b5f3bc/utils/funcs.py#L124
    """
    if type(lst) != np.ndarray:
        array = np.array(lst)
    else:
        array = lst
        
    org_array = np.copy(array)   # The original label/pred lst
    
    num_tracts = len(ordered_tract_cluster_mapping_dict.keys())  # anatomically plausible tracts + 1 other tract 43
    
    # convert cluster idx to tract idx. Iterate tracts then clusters
    for idx_tract, tract_name in enumerate(ordered_tract_cluster_mapping_dict.keys()):
        cur_clusters_names = ordered_tract_cluster_mapping_dict[tract_name]
        for cluster_name in cur_clusters_names:
            cluster_idx = int(cluster_name.split('_')[1][2:])-1   # '00031'->30,'00322'->321.  Don't use strip('0'), it will strip 0 in '80' as well.
            array[org_array==cluster_idx] = idx_tract  # plausible tract, starts from 0
            array[org_array==cluster_idx+800] = num_tracts-1  # implausible tract (outlier)
    
    if output_lst:   # when array is 1D
        new_lst = list(array)
    else:           # We could get array with 2D/3D/nD shape, e.g., (num_fiber_per_brian, num_points_per_fiber)
        new_lst = array
    
    return new_lst



def obtain_TractClusterMapping(input_path: Union[str, os.PathLike]) -> Dict[str, List[str]]:
    """
        The obtained dictionary is {'tract name': ['cluster_xxx','cluster_xxx', ... 'cluster_xxx']}. Cluster index starts from 1.

        This function is copied from TractCloud
        https://github.com/SlicerDMRI/TractCloud/blob/2c2869cde12b8d617d21be8e19fd33ae73b5f3bc/utils/funcs.py#L99
    """
    ordered_tract_names = list(obtain_TractFullName().keys())
    ordered_tract_cluster_mapping_dict = {}   # keep the order we like association, projection, commissural, cerebellar, superficial 
    
    cluster_annotation_pd = pd.read_excel(Path(input_path)/'FiberClusterAnnotation_Updated20230110.xlsx')
    cluster_tract_mapping = {clu:tra for clu, tra in zip(cluster_annotation_pd['Cluster Index'], cluster_annotation_pd['Final'])}

    # only include clusters into the anatomical tracts
    tract_cluster_mapping_dict = defaultdict(list)  # {'tract_x': ['cluster_xxx','cluster_xxx']}
    for key, value in cluster_tract_mapping.items():
        tract_cluster_mapping_dict[value].append(key)   
    
    # sort the tract order in the dictionary and add clusters belong to other tracts
    all_clusters_lst = list(cluster_annotation_pd['Cluster Index'])
    clusters_belong_anatomical_tracts_lst = []
    for tract_name in ordered_tract_names:
        ordered_tract_cluster_mapping_dict[tract_name] = tract_cluster_mapping_dict[tract_name]
        clusters_belong_anatomical_tracts_lst.extend(tract_cluster_mapping_dict[tract_name])
    ordered_tract_cluster_mapping_dict['Other'] = sorted(list(set(all_clusters_lst)-set(clusters_belong_anatomical_tracts_lst))) 
    
    return ordered_tract_cluster_mapping_dict



def obtain_TractFullName() -> Dict[str, str]: 
    """
        Returns a dict of the full tract names.

        This function is copied from TractCloud: 
        https://github.com/SlicerDMRI/TractCloud/blob/2c2869cde12b8d617d21be8e19fd33ae73b5f3bc/utils/funcs.py#L50

    """   
    tract_name_full_dict = {
        'AF': 'arcuate fasciculus',  # association 
        'CB': 'cingulum bundle', 
        'EC': 'external capsule',
        'EmC': 'extreme capsule',
        'ILF': 'inferior longitudinal fasciculus', 
        'IOFF': 'inferior occipito-frontal fasciculus', 
        'MdLF': 'middle longitudinal fasciculus', 
        'SLF-I': 'superior longitudinal fasciculus I', 
        'SLF-II': 'superior longitudinal fasciculus II', 
        'SLF-III': 'superior longitudinal fasciculus III', 
        'UF': 'uncinate fasciculus', 
        'CST':'corticospinal tract',  # projection 
        'CR-F':'corona-radiata-frontal (excluding the CST)', 
        'CR-P':'corona-radiata-parietal (excluding the CST)', 
        'SF':'striato-frontal', 
        'SO':'striato-occipital', 
        'SP':'striato-parietal',
        'TF':'thalamo-frontal',
        'TO':'thalamo-occipital', 
        'TT': 'thalamo-temporal', 
        'TP': 'thalamo-parietal', 
        'PLIC':'posterior limb of internal capsule',
        'CC1': 'corpus callosum 1',    # Commissural 
        'CC2': 'corpus callosum 2',  
        'CC3': 'corpus callosum 3', 
        'CC4': 'corpus callosum 4', 
        'CC5': 'corpus callosum 5', 
        'CC6': 'corpus callosum 6', 
        'CC7': 'corpus callosum 7', 
        'CPC': 'cortico-ponto-cerebellar',     # cerebellar
        'ICP': 'inferior cerebellar peduncle', 
        'Intra-CBLM-I-P':'intracerebellar input and Purkinje tract', 
        'Intra-CBLM-PaT':'intracerebellar parallel tract',
        'MCP':'middle cerebellar peduncle', 
        'Sup-F':'superficial-frontal',  # superficial
        'Sup-FP':'superficial-frontal-parietal',
        'Sup-O':'superficial-occipital',
        'Sup-OT':'superficial-occipital-temporal',
        'Sup-P':'superficial-parietal',
        'Sup-PO':'superficial-parietal-occipital',
        'Sup-PT':'superficial-parietal-temporal',
        'Sup-T':'superficial-temporal'
        }
    
    return tract_name_full_dict



if __name__ == '__main__':
    context_size = 2000
    inputPath="tractCloud_Dataset/TrainData_800clu800ol"
    train_ret= get_TractCloud_train_returner(inputPath=inputPath,
                            device=torch.device("cpu"))
    print(f"Shape of first train tractogram: {train_ret(torch.tensor(0))[0].shape}")


import torch.distributed as dist
from torch.nn.parallel import DistributedDataParallel as DDP
import os


def ddp_setup():
    """
        Function to setup the Distributed Data Parallel (DDP) Environment
    """
    dist.init_process_group(backend="nccl")



def ddp_cleanup():
    """
        Function to cleanup the Distributed Data Parallel (DDP) Environment
    """
    dist.destroy_process_group()


def ddp_is_running():
    if "RANK" in os.environ: 
        return True
    if dist.is_available() and dist.is_initialized(): 
        return True
    return False



import argparse
from .test import test
import torch

def main():
    parser = argparse.ArgumentParser(prog="rapidparc-test", description="Program to benchmark RapidParc on the test split of the TractCloud dataset")
    parser.add_argument("--model_name_or_path", type=str, required=True, help="Name of the Model, i.e. `rapidparc`, `rapidparc_v8` or `hemiaug`. Alternatively path to the experiment directory, which should contain the trained model and args.yml")
    parser.add_argument("--applyTestSetAugmentations", action="store_true", help="Whether to apply test set augmentations")
    parser.add_argument("--eval_batch_size", type=int, default=512, help="Evaluation batch size")
    parser.add_argument("--eval_context_size", type=int, default=2000, help="Evaluation context size")
    parser.add_argument("--device", type=str, default=None, help="Device to use (e.g., 'cuda', 'cpu')")
    parser.add_argument("--print_classification_report", action="store_true", help="Whether to print the classification report")
    parser.add_argument("--tractcloud_licence_consent_given", action="store_true", help="Whether you do agree to the slicer licence.")
    args = parser.parse_args()
    
    test(model_name_or_path=args.model_name_or_path,
         applyTestSetAugmentations=args.applyTestSetAugmentations,
         eval_batch_size=args.eval_batch_size,
         eval_context_size=args.eval_context_size,
         device=torch.device(args.device) if args.device is not None else None,
         print_classification_report=args.print_classification_report,
         tractcloud_licence_consent_given=args.tractcloud_licence_consent_given
    )


from typing import Union, Tuple
from pathlib import Path
import torch
import torch.nn as nn
import os


def get_latest_snapshot(snapshot_path: Union[str, os.PathLike]) -> Union[os.PathLike, None]:
    """
        Function to get the latest snapshot file name

        Args:
            snapshot_path (Union[str, Path]): Path to the snapshot directory

        Returns:
            Union[Path, None]: Path to the latest snapshot file
    """
    snapshots = list(Path(snapshot_path).glob("snapshot*.pt"))
    if not snapshots:
        print(f"No snapshots found in {snapshot_path}")
        return None
    if len(snapshots) == 1 and snapshots[0].name == "snapshot.pt":
        return snapshots[0]
    return max(snapshots, key=lambda x: int(x.stem.split("_")[-1]) if "_" in x.stem else -1)



def load_snapshot(snapshot_file_path: Union[str, Path],
                  model: nn.Module,
                  scheduler: torch.optim.lr_scheduler.LRScheduler,
                  optimizer: torch.optim.Optimizer):
    """
        Function to load a snapshot of the model, optimizer and scheduler
    """
    checkpoint = torch.load(snapshot_file_path, weights_only=True)
    model.load_state_dict(checkpoint['model'])
    epoch = checkpoint['epoch'] + 1
    scheduler.load_state_dict(checkpoint['scheduler'])
    optimizer.load_state_dict(checkpoint['optimizer'])
    return model, epoch, scheduler, optimizer



def store_snapshot(snapshot_file_path: Union[str, Path],
                   model: nn.Module,
                   epoch: int,
                   scheduler: torch.optim.lr_scheduler.LRScheduler,
                   optimizer: torch.optim.Optimizer) -> None:
    """
        Function to store a snapshot of the model, optimizer and scheduler
    """
    checkpoint = {
        'model': model.state_dict(),
        'epoch': epoch,
        'scheduler': scheduler.state_dict(),
        'optimizer': optimizer.state_dict()
    }
    torch.save(checkpoint, snapshot_file_path)



def load_only_model_snapshot(snapshot_path: Union[str, Path],
                  model: nn.Module) -> nn.Module:
    """
        Function to load a snapshot of the model

        Args:
            snapshot_file_path (Union[str, Path]): Path to the snapshot file
            model (nn.Module): Model to be loaded

        Returns:
            nn.Module: Model with loaded weights
    """
    checkpoint = torch.load(snapshot_path, weights_only=True)
    model.load_state_dict(checkpoint['model'])
    return model


import torch
import torch.nn as nn
from torch.utils import data
from tqdm import tqdm
import numpy as np
from typing import Union, Tuple, List, Callable, Optional
from pathlib import Path
import torch.distributed as dist
from torch.utils.data.distributed import DistributedSampler

from .ddp_handling import ddp_is_running
from .model_snapshot_handling import get_latest_snapshot, load_snapshot, store_snapshot


def train(get_train_Tensor: Callable[[Union[List[int], torch.Tensor]], Tuple[torch.Tensor, torch.Tensor]],
          get_val_Tensor: Callable[[], Tuple[torch.Tensor, torch.Tensor, torch.Tensor]],
          num_epochs: int,
          batch_size: int,
          context_size: int, 
          optimizer: torch.optim.Optimizer,
          scheduler: torch.optim.lr_scheduler.LRScheduler,
          model: nn.Module,
          device: torch.device,
          criterion: nn.Module,
          snapshot_path: Union[str, Path],
          snapshot_every: int,
          store_all_snapshots: bool,
          map_800_800_to_43: torch.Tensor,
          transforms_on_gpu: nn.Module,
          aug_p_hemi: float = 0.0,
          aug_method_hemi: str = 'replace') -> nn.Module:
    """ 
    Train function to train the model
    
    Parameters:
        get_train_Tensor (Callable): Function to get the training data
        get_val_Tensor (Callable): Function to get the validation data
        num_epochs (int): Number of epochs to train
        batch_size (int): Batch size
        context_size (int): Context size
        optimizer (torch.optim.Optimizer): Optimizer
        scheduler (torch.optim.lr_scheduler): Scheduler
        model (nn.Module): Model
        device (torch.device): Device
        criterion (nn.Module): Criterion
        snapshot_path (Union[str, Path]): Path to store the snapshots
        snapshot_every (int): Store a snapshot every n epochs
        store_all_snapshots (bool): Store all snapshots
        map_800_800_to_43 (torch.tensor): Mapping from 800x800 labels to 43 labels
        transforms_on_gpu (nn.ModuleList): List of transformations to be applied on the GPU

    Returns:
        nn.Module: Trained model
    """
    if ddp_is_running():
        dist_sampler = DistributedSampler(dataset=torch.utils.data.TensorDataset(torch.arange(70)), 
                                            num_replicas=dist.get_world_size(), 
                                            rank=dist.get_rank(), 
                                            shuffle=True)
    else:
        dist_sampler = torch.arange(70)
    snapshot_path = Path(snapshot_path)
    map_800_800_to_43 = map_800_800_to_43.to(device)

    ####### Load checkpoints if needed #######
    start_epoch = 0
    start_snapshot = get_latest_snapshot(snapshot_path)
    if start_snapshot is not None:
        if dist.get_rank() == 0: print(f"Loading Snapshot {start_snapshot}")
        model, start_epoch, scheduler, optimizer = load_snapshot(snapshot_file_path=start_snapshot, model=model, scheduler=scheduler, optimizer=optimizer)


    if not ddp_is_running() or dist.get_rank() == 0: 
        print(f"Starting from epoch {start_epoch}")
        epoch_range = tqdm(range(start_epoch, num_epochs))
    else: 
        epoch_range = range(start_epoch, num_epochs)

    for epoch in epoch_range:
        ######## validation epoch ########
        if not ddp_is_running() or dist.get_rank() == 0:
            # Evaluate the model only on rank 0
            accuracy_43, accuracy_800_800, val_loss = eval_model(get_val_Tensor=get_val_Tensor, 
                                        model=model, 
                                        criterion=criterion,
                                        map_800_800_to_43=map_800_800_to_43)

        ######## training epoch ########
        cur_loss_iters = train_epoch(model=model,
                                     epoch=epoch,
                                     batch_size=batch_size,
                                     context_size=context_size,
                                     get_train_Tensor=get_train_Tensor,
                                     optimizer=optimizer,
                                     scheduler=scheduler,
                                     criterion=criterion,
                                     transforms_on_gpu=transforms_on_gpu,
                                     aug_p_hemi=aug_p_hemi,
                                     aug_method_hemi=aug_method_hemi,
                                     device=device,
                                     distributed_sampler=dist_sampler)

        if ddp_is_running() and dist.get_world_size() > 1:
            dist.all_reduce(cur_loss_iters, op=dist.ReduceOp.SUM)
            cur_loss_iters = cur_loss_iters.cpu() / dist.get_world_size()
            mean_loss = cur_loss_iters.mean()
        else:
            cur_loss_iters = cur_loss_iters.cpu()
            mean_loss = cur_loss_iters.mean()

        ######## logging ########
        if not ddp_is_running() or dist.get_rank() == 0:

            ######## Checkpoint ######## # 
            if (epoch + 1) % snapshot_every == 0 or epoch == num_epochs - 1:
                print(f"Epoch {epoch+1}/{num_epochs}")
                print(f"    Train loss: {mean_loss}")
                # print(f"    Valid loss: {val_loss}")
                # print(f"    Valid accuracy 43: {accuracy_43}")
                # print(f"    Valid accuracy 800_800: {accuracy_800_800}")
                print("\n")
                snapshot_file_path = snapshot_path / f"snapshot_{epoch+1}.pt" if store_all_snapshots else snapshot_path / "snapshot.pt"
                store_snapshot(snapshot_file_path, model, epoch, scheduler, optimizer)
                print(f"Snapshot stored at {snapshot_file_path}")

    return model



def train_epoch(model: nn.Module,
                epoch: int,
                batch_size: int,
                get_train_Tensor: Callable[[Union[List[int], torch.Tensor]], Tuple[torch.Tensor, torch.Tensor]],
                optimizer: torch.optim.Optimizer,
                criterion: nn.Module,
                scheduler: torch.optim.lr_scheduler.LRScheduler,
                distributed_sampler: Union[DistributedSampler, torch.Tensor],
                context_size: int,
                transforms_on_gpu: nn.Module,
                aug_p_hemi: float,
                aug_method_hemi: str,
                device: Union[int, torch.device]
                ) -> torch.Tensor:
    """ 
    Training the model for one epoch
    Parameters:
        model (nn.Module): Model
        epoch (int): Epoch
        batch_size (int): Batch size
        get_train_Tensor (Callable): Function to get the training data
        optimizer (torch.optim.Optimizer): Optimizer
        criterion (nn.Module): Criterion
        scheduler (torch.optim.lr_scheduler): Scheduler
        distributed_sampler (DistributedSampler | torch.Tensor): Distributed Sampler
        context_size (int): Context size
        transforms_on_gpu (nn.ModuleList): List of transformations to be applied on the GPU
        device (Union[int, torch.device]): Device

    Returns:
        torch.tensor: Loss list
    """
    with torch.no_grad():
        if ddp_is_running():
            distributed_sampler.set_epoch(epoch)
        data, labels_800_800 = get_train_Tensor(list(distributed_sampler))

        # Random subtractogram-partitioning: 
        # -> Shuffle Streamlines for each Subject, bring in ContextSize_bits
        num_subjects, num_streamlines, num_points, stml_dim = data.shape
        perms = torch.stack([torch.randperm(num_streamlines, device=device) for _ in range(num_subjects)])
        data = data[torch.arange(num_subjects).unsqueeze(-1), perms]
        labels_800_800 = labels_800_800[torch.arange(num_subjects).unsqueeze(-1), perms] 
    
        # Single hemisphere augmentation
        if aug_p_hemi == 0.0:
            data = transforms_on_gpu(data.reshape(-1, context_size, num_points, stml_dim))
        else:
            data, labels_800_800 = _apply_hemi_aug(data, labels_800_800, aug_p_hemi, aug_method_hemi, context_size)
            data = transforms_on_gpu(data)

        labels_800_800 = labels_800_800.reshape(-1, context_size) # [X, context_size] with X = num_subjects * num_streamlines / context_size

        # Shuffle data, note that this operation can be done on every rank independently and we do not need a distributed sampler
        # since the data is already distributed (using a distributed sampler inside the get_train_Tensor function)
        indices = torch.randperm(data.shape[0])
        data = data[indices]
        labels_800_800 = labels_800_800[indices]


    loss_list = torch.zeros((int(np.ceil(len(data) / batch_size))), device=device, requires_grad=False)
    
    model.train()
    for iterIndex, i in enumerate(torch.arange(0, len(data), batch_size)):
        data_batch = data[i:i+batch_size].requires_grad_() # [bs, context_size, num_points, stml_dim]
        labels_800_800_batch = labels_800_800[i:i+batch_size] # [bs, context_size]

        # Clear gradients w.r.t. parameters
        optimizer.zero_grad(set_to_none=True)

        # Forward pass to get output/logits
        outputs = model(data_batch)

        # Calculate Loss: softmax --> cross entropy loss
        # Transpose to have dim_out as 1st dimension (required by CrossEntropyLoss)
        loss = criterion(torch.transpose(outputs, 1, 2), labels_800_800_batch)
        loss_list[iterIndex] = loss

        # Updating parameters
        loss.backward()
        optimizer.step()
    scheduler.step()

    return loss_list



@torch.no_grad()
def _apply_hemi_aug(data, labels_800_800, aug_p_hemi, aug_method_hemi, context_size):
    _, num_streamlines, num_points, stml_dim = data.shape
    data_subjectwise = data.clone()  # before reshaping to [B, context_size, ...]
    labels_subjectwise = labels_800_800.clone()
    
    data = data.reshape(-1, context_size, num_points, stml_dim)        # [B, C, P, D]
    labels_800_800 = labels_800_800.reshape(-1, context_size)          # [B, C]

    for i in range(data.shape[0]):
        if torch.rand(1).item() < aug_p_hemi:
            subtractogram = data[i]                  # [C, P, D]
            centroids = subtractogram.mean(dim=1)    # [C, D]
            x_coords = centroids[:, 0]

            left_mask = x_coords < 0
            right_mask = x_coords >= 0

            assert aug_method_hemi == 'random', 'Only random augmentation is implemented for hemi-augmentation.'

            if aug_method_hemi == 'random':
                subj_id = i // (num_streamlines // context_size)

                subj_data = data_subjectwise[subj_id]           # [N, P, D]
                subj_labels = labels_subjectwise[subj_id]       # [N]

                centroids = subj_data.mean(dim=1)               # [N, D]
                x_coords = centroids[:, 0]
                left_mask = x_coords < 0
                right_mask = x_coords >= 0

                mode = torch.randint(0, 3, (1,), device=data.device).item()  # 0 = left, 1 = right, 2 = both

                if mode == 0 and left_mask.sum() > 0:
                    pool_data = subj_data[left_mask]
                    pool_labels = subj_labels[left_mask]
                elif mode == 1 and right_mask.sum() > 0:
                    pool_data = subj_data[right_mask]
                    pool_labels = subj_labels[right_mask]
                else:
                    pool_data = subj_data
                    pool_labels = subj_labels

                pool_size = pool_data.shape[0]
                sample_idx = torch.randint(0, pool_size, (context_size,), device=data.device)

                data[i] = pool_data[sample_idx]
                labels_800_800[i] = pool_labels[sample_idx]
    
    return data, labels_800_800




@torch.inference_mode()
def eval_model(get_val_Tensor: Callable, 
               model: nn.Module, 
               criterion: nn.Module, 
               map_800_800_to_43: torch.Tensor) -> Tuple[float, float, float]:
    """ Evaluating the model for one epoch

    Parameters:
        get_val_Tensor (Callable): Function to get the validation data
        model (nn.Module): Model
        criterion (nn.Module): Criterion
        map_800_800_to_43 (torch.tensor): Mapping from 800x800 labels to 43 labels

    Returns:
        Tuple[float, float, float]: Accuracy_43, Accuracy_800_800, Loss
    """
    correct_800_800 = 0
    correct_43 = 0

    model.eval()
    data, labels_43, labels_800_800 = get_val_Tensor()

    # Forward pass only to get logits/output
    outputs = model(data) # Shape [bs, dim_out, context_size, dim_out]

    # Transpose to have dim_out as 1st dimension (required by CrossEntropyLoss)
    loss = criterion(torch.transpose(outputs, 1, 2), labels_800_800).item()

    # Get predictions from the maximum value
    preds = torch.argmax(outputs, dim = -1) # [bs, contextSize]
    correct_800_800 = len( torch.where(preds==labels_800_800)[0])
    correct_43 = len( torch.where(map_800_800_to_43[preds]==labels_43)[0])

    total = labels_43.shape[0] * labels_43.shape[1]

    # Total correct predictions and loss
    accuracy_800_800 = 100 * correct_800_800 / total
    accuracy_43 = 100 * correct_43 / total

    return accuracy_43, accuracy_800_800, loss
